# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from io import StringIO
import pytest
from unittest import TestCase

import montepy
from montepy.errors import *
from montepy.input_parser import input_syntax_reader
from montepy.input_parser.mcnp_input import Input, Jump, Message, ReadInput, Title
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.input_file import MCNP_InputFile
from montepy.input_parser.parser_base import MCNP_Parser
from montepy.input_parser.shortcuts import Shortcuts
from montepy.input_parser import syntax_node
from montepy.particle import Particle
import warnings


class TestValueNode(TestCase):
    def test_valuenoode_init(self):
        for type, token, answer in [
            (str, "hi", "hi"),
            (float, "1.2300", 1.23),
            (int, "1", 1),
            (float, "1.23e-3", 1.23e-3),
            (float, "6.02+23", 6.02e23),
        ]:
            for padding in [None, syntax_node.PaddingNode(" ")]:
                node = syntax_node.ValueNode(token, type, padding)
                self.assertEqual(node.value, answer)
                self.assertEqual(node.token, token)
                if padding:
                    self.assertEqual(node.padding, padding)
                else:
                    self.assertIsNone(node.padding)
        # test with None values
        for type in {str, float, int}:
            node = syntax_node.ValueNode(None, type)
            self.assertIsNone(node.value)
            node = syntax_node.ValueNode(Jump(), type)
            self.assertIsNone(node.value)

    def test_valuenode_convert_to_int(self):
        node = syntax_node.ValueNode("1", float)
        node._convert_to_int()
        self.assertEqual(node.type, int)
        self.assertEqual(node.value, 1)
        # test 1.0
        node = syntax_node.ValueNode("1.0", float)
        node._convert_to_int()
        self.assertEqual(node.type, int)
        self.assertEqual(node.value, 1)
        # test wrong type
        with self.assertRaises(ValueError):
            node = syntax_node.ValueNode("hi", str)
            node._convert_to_int()
        # test real float
        with self.assertRaises(ValueError):
            node = syntax_node.ValueNode("1.23", float)
            node._convert_to_int()

    def test_valuenode_convert_to_enum(self):
        node = syntax_node.ValueNode("1", float)
        lat = montepy.data_inputs.lattice.Lattice
        node._convert_to_enum(lat)
        self.assertEqual(node.type, lat)
        self.assertEqual(node.value, lat(1))
        # test with None
        with self.assertRaises(ValueError):
            node = syntax_node.ValueNode(None, float)
            node._convert_to_enum(lat)
        node._convert_to_enum(lat, allow_none=True)
        self.assertIsNone(node.value)
        st = montepy.surfaces.surface_type.SurfaceType
        node = syntax_node.ValueNode("p", str)
        node._convert_to_enum(st, switch_to_upper=True)
        self.assertEqual(node.type, st)
        self.assertEqual(node.value, st("P"))

    def test_is_negat_identifier(self):
        node = syntax_node.ValueNode("-1", float)
        self.assertTrue(not node.is_negatable_identifier)
        self.assertIsNone(node.is_negative)
        node.is_negatable_identifier = True
        self.assertTrue(node.is_negatable_identifier)
        self.assertEqual(node.type, int)
        self.assertTrue(node.value > 0)
        self.assertTrue(node.is_negative)
        # test with positive number
        node = syntax_node.ValueNode("1", float)
        node.is_negatable_identifier = True
        self.assertEqual(node.type, int)
        self.assertTrue(node.value > 0)
        self.assertTrue(not node.is_negative)
        # test with none
        node = syntax_node.ValueNode(None, float)
        node.is_negatable_identifier = True
        self.assertEqual(node.type, int)
        self.assertIsNone(node.value)
        self.assertIsNone(node.is_negative)
        node.value = 1
        self.assertEqual(node.value, 1)
        self.assertTrue(not node.is_negative)

    def test_is_negat_float(self):
        node = syntax_node.ValueNode("-1.23", float)
        self.assertTrue(not node.is_negatable_float)
        self.assertIsNone(node.is_negative)
        node.is_negatable_float = True
        self.assertEqual(node.type, float)
        self.assertTrue(node.value > 0)
        self.assertTrue(node.is_negative)
        self.assertTrue(node.is_negatable_float)
        # test with positive number
        node = syntax_node.ValueNode("1.23", float)
        node.is_negatable_float = True
        self.assertEqual(node.type, float)
        self.assertTrue(not node.is_negative)
        # test with None
        node = syntax_node.ValueNode(None, float)
        node.is_negatable_float = True
        self.assertEqual(node.type, float)
        self.assertIsNone(node.value)
        self.assertIsNone(node.is_negative)
        node.value = 1
        self.assertEqual(node.value, 1)
        self.assertTrue(not node.is_negative)

    def test_is_negative(self):
        node = syntax_node.ValueNode("-1.23", float)
        node.is_negatable_float = True
        self.assertTrue(node.is_negative)
        node.is_negative = False
        self.assertTrue(node.value > 0)
        self.assertTrue(not node.is_negative)
        node = syntax_node.ValueNode("hi", str)
        node.is_negative = True
        self.assertIsNone(node.is_negative)

    def test_valuenode_int_format(self):
        node = syntax_node.ValueNode("-1", int)
        answer = "-1"
        output = node.format()
        self.assertEqual(output, answer)
        for input, val, answer, expand in [
            ("1", 5, "5", False),
            (1, 5, "5", False),
            ("-1", 2, " 2", False),
            ("-1", -2, "-2", False),
            ("+1", 5, "+5", False),
            ("0001", 5, "0005", False),
            (Jump(), 5, "5 ", False),
        ]:
            print("in:", input, "new value:", val, "answer:", answer)
            node = syntax_node.ValueNode(input, int)
            node.value = val
            self.assertEqual(node.format(), answer)
            node = syntax_node.ValueNode(input, int)
            node.is_negatable_identifier = True
            node.value = val
            node.is_negative = val < 0
            if expand:
                with self.assertWarns(LineExpansionWarning):
                    self.assertEqual(node.format(), answer)
            else:
                self.assertEqual(node.format(), answer)
        # test messing around with padding
        for padding, val, answer, expand in [
            ([" "], 10, "10 ", True),
            (["  "], 10, "10 ", False),
            (["\n"], 10, "10\n", True),
            ([" ", "\n", "c hi"], 10, "10\nc hi", False),
            (["  ", "\n"], 10, "10 \n", False),
        ]:
            print("padding", padding, "new_val", val, "answer", repr(answer))
            pad_node = syntax_node.PaddingNode(padding[0])
            for pad in padding[1:]:
                pad_node.append(pad)
            node = syntax_node.ValueNode("1", int, pad_node)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with self.assertWarns(LineExpansionWarning):
                    self.assertEqual(node.format(), answer)
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                self.assertEqual(node.format(), answer)

    def test_value_has_changed(self):
        # test None no change
        node = syntax_node.ValueNode(None, int)
        self.assertTrue(not node._value_changed)
        # test None changed
        node.value = 5
        self.assertTrue(node._value_changed)
        node = syntax_node.ValueNode("1.23", float)
        self.assertTrue(not node._value_changed)
        node.value = 1.25
        self.assertTrue(node._value_changed)
        node = syntax_node.ValueNode("hi", str)
        self.assertTrue(not node._value_changed)
        node.value = "foo"
        self.assertTrue(node._value_changed)

    def test_value_float_format(self):
        for input, val, answer, expand in [
            ("1.23", 1.23, "1.23", False),
            ("1.23", 4.56, "4.56", False),
            (1.23, 4.56, "4.56", False),
            ("-1.23", 4.56, " 4.56", False),
            ("1.0e-2", 2, "2.0e+0", False),
            ("1.602-19", 6.02e23, "6.020+23", False),
            ("1.602-0019", 6.02e23, "6.020+0023", False),
            (Jump(), 5.4, "5.4 ", True),
            ("1", 2, "2", False),
            ("0.5", 0, "0.0", False),
        ]:
            print("in:", input, "new value:", val, "answer:", answer)
            node = syntax_node.ValueNode(input, float)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with self.assertWarns(LineExpansionWarning):
                    self.assertEqual(node.format(), answer)
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                self.assertEqual(node.format(), answer)
        for padding, val, answer, expand in [
            ([" "], 10, "10.0 ", True),
            (["  "], 10, "10.0 ", False),
            (["\n"], 10, "10.0\n", True),
            ([" ", "\n", "c hi"], 10, "10.0\nc hi", False),
            (["  ", "\n"], 10, "10.0 \n", False),
        ]:
            print("padding", padding, "new_val", val, "answer", answer)
            pad_node = syntax_node.PaddingNode(padding[0])
            for pad in padding[1:]:
                pad_node.append(pad)
            node = syntax_node.ValueNode("1.0", float, pad_node)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with self.assertWarns(LineExpansionWarning):
                    self.assertEqual(node.format(), answer)
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                self.assertEqual(node.format(), answer)

    def test_value_str_format(self):
        for input, val, answer, expand in [
            ("hi", "foo", "foo", True),
            ("hi", None, "", False),
        ]:
            node = syntax_node.ValueNode(input, str)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with self.assertWarns(LineExpansionWarning):
                    self.assertEqual(node.format(), answer)
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                self.assertEqual(node.format(), answer)
        for padding, val, answer, expand in [
            ([" "], "foo", "foo ", True),
            (["  "], "foo", "foo ", False),
            (["\n"], "foo", "foo\n", True),
            ([" ", "\n", "c hi"], "foo", "foo\nc hi", False),
            (["  ", "\n"], "foo", "foo \n", False),
        ]:
            print("padding", padding, "new_val", val, "answer", answer)
            pad_node = syntax_node.PaddingNode(padding[0])
            for pad in padding[1:]:
                pad_node.append(pad)
            node = syntax_node.ValueNode("hi", str, pad_node)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with self.assertWarns(LineExpansionWarning):
                    self.assertEqual(node.format(), answer)
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                self.assertEqual(node.format(), answer)

    def test_value_enum_format(self):
        lat = montepy.data_inputs.lattice.Lattice
        st = montepy.surfaces.surface_type.SurfaceType
        for input, val, enum_class, args, answer, expand in [
            (
                "1",
                lat.HEXAGONAL,
                lat,
                {"format_type": int, "switch_to_upper": False},
                "2",
                False,
            ),
            (
                1,
                lat.HEXAGONAL,
                lat,
                {"format_type": int, "switch_to_upper": False},
                "2",
                False,
            ),
            ("p", st.PZ, st, {"format_type": str, "switch_to_upper": True}, "PZ", True),
        ]:
            node = syntax_node.ValueNode(input, args["format_type"])
            node._convert_to_enum(enum_class, **args)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with self.assertWarns(LineExpansionWarning):
                    self.assertEqual(node.format(), answer)
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                self.assertEqual(node.format(), answer)

    def test_value_comments(self):
        value_node = syntax_node.ValueNode("1", int)
        self.assertEqual(len(list(value_node.comments)), 0)
        padding = syntax_node.PaddingNode("$ hi", True)
        value_node.padding = padding
        comments = list(value_node.comments)
        self.assertEqual(len(comments), 1)
        self.assertIn("hi", comments[0].contents)

    def test_value_trailing_comments(self):
        value_node = syntax_node.ValueNode("1", int)
        self.assertIsNone(value_node.get_trailing_comment())
        value_node._delete_trailing_comment()
        self.assertIsNone(value_node.get_trailing_comment())
        padding = syntax_node.PaddingNode("c hi", True)
        value_node.padding = padding
        comment = value_node.get_trailing_comment()
        self.assertEqual(len(comment), 1)
        self.assertEqual(comment[0].contents, "hi")
        value_node._delete_trailing_comment()
        self.assertIsNone(value_node.get_trailing_comment())

    def test_value_str(self):
        value_node = syntax_node.ValueNode("1", int)
        str(value_node)
        repr(value_node)
        padding = syntax_node.PaddingNode("$ hi", True)
        value_node.padding = padding
        str(value_node)
        repr(value_node)

    def test_value_equality(self):
        value_node1 = syntax_node.ValueNode("1", int)
        self.assertTrue(value_node1 == value_node1)
        with self.assertRaises(TypeError):
            value_node1 == syntax_node.PaddingNode("")
        value_node2 = syntax_node.ValueNode("2", int)
        self.assertTrue(value_node1 != value_node2)
        value_node3 = syntax_node.ValueNode("hi", str)
        self.assertTrue(value_node1 != value_node3)
        self.assertTrue(value_node1 == 1)
        self.assertTrue(value_node1 != 2)
        self.assertTrue(value_node1 != "hi")
        value_node4 = syntax_node.ValueNode("1.5", float)
        value_node5 = syntax_node.ValueNode("1.50000000000001", float)
        self.assertTrue(value_node4 == value_node5)
        value_node5.value = 2.0
        self.assertTrue(value_node4 != value_node5)


class TestSyntaxNode(TestCase):
    def setUp(self):
        value1 = syntax_node.ValueNode("1.5", float)
        value2 = syntax_node.ValueNode("1", int)
        self.test_node = syntax_node.SyntaxNode(
            "test",
            {"foo": value1, "bar": value2, "bar2": syntax_node.SyntaxNode("test2", {})},
        )

    def test_syntax_init(self):
        test = self.test_node
        self.assertEqual(test.name, "test")
        self.assertIn("foo", test.nodes)
        self.assertIn("bar", test.nodes)
        self.assertIsInstance(test.nodes["foo"], syntax_node.ValueNode)

    def test_syntax_name(self):
        test = self.test_node
        test.name = "hi"
        self.assertEqual(test.name, "hi")
        with self.assertRaises(TypeError):
            test.name = 1.0

    def test_get_value(self):
        test = self.test_node
        self.assertEqual(test.get_value("foo"), 1.5)
        with self.assertRaises(KeyError):
            test.get_value("foo2")
        with self.assertRaises(KeyError):
            test.get_value("bar2")

    def test_syntax_format(self):
        output = self.test_node.format()
        self.assertEqual(output, "1.51")

    def test_syntax_dict(self):
        test = self.test_node
        self.assertIn("foo", test)
        self.assertEqual(test["foo"], test.nodes["foo"])

    def test_syntax_comments(self):
        padding = syntax_node.PaddingNode("$ hi", True)
        test = copy.deepcopy(self.test_node)
        test["foo"].padding = padding
        padding = syntax_node.PaddingNode("$ foo", True)
        test["bar"].padding = padding
        comments = list(test.comments)
        self.assertEqual(len(comments), 2)

    def test_syntax_trailing_comments(self):
        # test with blank tail
        self.assertIsNone(self.test_node.get_trailing_comment())
        test = copy.deepcopy(self.test_node)
        test["bar2"].nodes["foo"] = syntax_node.ValueNode("1.23", float)
        self.assertIsNone(test.get_trailing_comment())
        test["bar2"]["foo"].padding = syntax_node.PaddingNode("c hi", True)
        self.assertEqual(len(test.get_trailing_comment()), 1)
        test._delete_trailing_comment()
        self.assertIsNone(test.get_trailing_comment())

    def test_syntax_str(self):
        str(self.test_node)
        repr(self.test_node)


class TestGeometryTree(TestCase):
    def setUp(self):
        left = syntax_node.ValueNode("1", int)
        right = syntax_node.ValueNode("2", int)
        op = syntax_node.PaddingNode(" ")
        self.test_tree = syntax_node.GeometryTree(
            "test",
            {"left": left, "operator": op, "right": right},
            montepy.Operator.INTERSECTION,
            left,
            right,
        )

    def test_geometry_init(self):
        left = syntax_node.ValueNode("1", int)
        right = syntax_node.ValueNode("2", int)
        op = syntax_node.PaddingNode(" ")
        tree = syntax_node.GeometryTree(
            "test",
            {"left": left, "operator": op, "right": right},
            montepy.Operator.INTERSECTION,
            left,
            right,
        )
        self.assertIs(tree.left, left)
        self.assertIs(tree.right, right)
        self.assertEqual(tree.operator, montepy.Operator.INTERSECTION)

    def test_geometry_format(self):
        test = self.test_tree
        self.assertEqual(test.format(), "1 2")

    def test_geometry_str(self):
        test = self.test_tree
        str(test)
        repr(test)

    def test_geometry_comments(self):
        test = copy.deepcopy(self.test_tree)
        test.left.padding = syntax_node.PaddingNode("$ hi", True)
        comments = list(test.comments)
        self.assertEqual(len(comments), 1)


def test_geometry_tree_mark_last_leaf_shortcut():
    # test left leaf
    tree = {
        "left": syntax_node.ValueNode("123", int),
        "operator": syntax_node.PaddingNode(" "),
    }
    geom = syntax_node.GeometryTree(
        "test", tree, montepy.Operator.INTERSECTION, tree["left"]
    )
    geom.mark_last_leaf_shortcut(Shortcuts.REPEAT)
    assert geom._left_short_type == Shortcuts.REPEAT
    # try overriding old shortcut
    geom.mark_last_leaf_shortcut(Shortcuts.MULTIPLY)
    assert geom._left_short_type == Shortcuts.REPEAT
    tree["right"] = syntax_node.ValueNode("789", int)
    geom = syntax_node.GeometryTree(
        "test", tree, montepy.Operator.INTERSECTION, tree["left"], tree["right"]
    )
    del tree["right"]
    tree["left"] = syntax_node.ValueNode("456", int)
    # make unbalanced tree in the other way for rigor
    geom = syntax_node.GeometryTree(
        "test", tree, montepy.Operator.INTERSECTION, tree["left"], geom
    )
    # test a right-side leaf
    geom.mark_last_leaf_shortcut(Shortcuts.REPEAT)
    assert geom.right._right_short_type == Shortcuts.REPEAT
    # try overriding old shortcut
    geom.mark_last_leaf_shortcut(Shortcuts.MULTIPLY)
    assert geom.right._right_short_type == Shortcuts.REPEAT


class TestPaddingNode(TestCase):
    def test_padding_init(self):
        pad = syntax_node.PaddingNode(" ")
        self.assertEqual(len(pad.nodes), 1)
        self.assertEqual(pad.value, " ")

    def test_padding_is_space(self):
        pad = syntax_node.PaddingNode(" ")
        self.assertTrue(pad.is_space(0))
        pad.append("\n")
        self.assertTrue(not pad.is_space(1))
        pad.append("$ hi", True)
        self.assertTrue(not pad.is_space(2))
        with self.assertRaises(IndexError):
            pad.is_space(5)

    def test_padding_append(self):
        pad = syntax_node.PaddingNode(" ")
        pad.append("\n")
        self.assertEqual(len(pad), 2)
        pad.append(" ")
        self.assertEqual(len(pad), 3)
        pad.append(" \n")
        self.assertEqual(len(pad), 5)
        pad.append("$ hi", True)
        self.assertEqual(len(pad), 6)

    def test_padding_format(self):
        pad = syntax_node.PaddingNode(" ")
        self.assertEqual(pad.format(), " ")
        pad.append("$ hi", True)
        self.assertEqual(pad.format(), " $ hi")

    def test_padding_grab_beginning_format(self):
        pad = syntax_node.PaddingNode(" ")
        new_pad = [
            syntax_node.CommentNode("c hi"),
            "\n",
            syntax_node.CommentNode("c foo"),
        ]
        answer = copy.copy(new_pad)
        pad._grab_beginning_comment(new_pad)
        self.assertEqual(pad.nodes, answer + ["\n", " "])

    def test_padding_eq(self):
        pad = syntax_node.PaddingNode(" ")
        self.assertTrue(pad == " ")
        self.assertTrue(pad != " hi ")
        pad1 = syntax_node.PaddingNode(" ")
        self.assertTrue(pad == pad1)
        with self.assertRaises(TypeError):
            pad == 1

    def test_comment_init(self):
        comment = syntax_node.CommentNode("$ hi")
        self.assertIsInstance(comment.nodes[0], syntax_node.SyntaxNode)
        self.assertEqual(len(comment.nodes), 1)
        self.assertTrue(comment.is_dollar)
        comment = syntax_node.CommentNode(" c hi")
        self.assertTrue(not comment.is_dollar)
        self.assertEqual(len(list(comment.comments)), 1)

    def test_comment_append(self):
        comment = syntax_node.CommentNode("c foo")
        comment.append("c bar")
        self.assertEqual(len(comment.nodes), 2)
        # test mismatch
        comment = syntax_node.CommentNode("$ hi")
        with self.assertRaises(TypeError):
            comment.append("c hi")

    def test_comment_str(self):
        comment = syntax_node.CommentNode("$ hi")
        str(comment)
        repr(comment)

    def test_blank_dollar_comment(self):
        comment = syntax_node.CommentNode("$")
        self.assertTrue(comment.is_dollar)
        self.assertEqual(len(list(comment.comments)), 1)
        self.assertEqual(len(comment.contents), 0)


def test_graveyard_comment():
    padding = syntax_node.PaddingNode(" ")
    padding.append("$ test", True)
    assert padding.has_graveyard_comment()
    padding.append(" ")
    assert padding.has_graveyard_comment()
    padding.append("\n")
    assert not padding.has_graveyard_comment()


@pytest.mark.parametrize(
    "padding,expect",
    [
        (["$ hi"], False),
        ([" c style comment"], False),
        ([" "], True),
        (["$ hi", "    ", "c hi"], True),
    ],
)
def test_padding_has_space(padding, expect):
    node = syntax_node.PaddingNode(padding[0])
    for pad in padding[1:]:
        node.append(pad)
    assert node.has_space() == expect


class TestParticlesNode(TestCase):
    def test_particle_init(self):
        parts = syntax_node.ParticleNode("test", ":n,p,e")
        particle = montepy.particle.Particle
        answers = {particle.NEUTRON, particle.PHOTON, particle.ELECTRON}
        self.assertEqual(parts.particles, answers)
        self.assertEqual(len(list(parts.comments)), 0)
        for part in parts:
            self.assertIn(part, answers)
        answers = [particle.NEUTRON, particle.PHOTON, particle.ELECTRON]
        self.assertEqual(parts._particles_sorted, answers)

    def test_particles_setter(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        particle = montepy.particle.Particle
        parts.particles = {particle.TRITON}
        self.assertEqual(parts.particles, {particle.TRITON})
        parts.particles = [particle.TRITON]
        self.assertEqual(parts.particles, {particle.TRITON})
        with self.assertRaises(TypeError):
            parts.particles = "hi"
        with self.assertRaises(TypeError):
            parts.particles = {"hi"}

    def test_particles_add_remove(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        particle = montepy.particle.Particle
        parts.add(particle.TRITON)
        self.assertIn(particle.TRITON, parts)
        self.assertEqual(parts._particles_sorted[-1], particle.TRITON)
        with self.assertRaises(TypeError):
            parts.add("hi")
        parts.remove(particle.NEUTRON)
        self.assertNotIn(particle.NEUTRON, parts)
        with self.assertRaises(TypeError):
            parts.remove("hi")

    def test_particles_sorted(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        particle = montepy.particle.Particle
        # lazily work around internals
        parts._particles.remove(particle.NEUTRON)
        self.assertNotIn(particle.NEUTRON, parts._particles_sorted)
        parts._particles.add(particle.TRITON)
        self.assertIn(particle.TRITON, parts._particles_sorted)

    def test_particles_format(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        repr(parts)
        self.assertEqual(parts.format(), ":n,p,e")
        parts = syntax_node.ParticleNode("test", "N,P,E")
        self.assertEqual(parts.format(), ":N,P,E")


class TestListNode(TestCase):
    def test_list_init(self):
        list_node = syntax_node.ListNode("list")
        self.assertEqual(list_node.nodes, [])

    def test_list_append(self):
        list_node = syntax_node.ListNode("list")
        list_node.append(syntax_node.ValueNode("1.0", float))
        self.assertEqual(len(list_node), 1)

    def test_list_str(self):
        list_node = syntax_node.ListNode("list")
        list_node.append(syntax_node.ValueNode("1.0", float))
        str(list_node)
        repr(list_node)

    def test_list_slicing(self):
        list_node = syntax_node.ListNode("list")
        for i in range(20):
            list_node.append(syntax_node.ValueNode("1.0", float))
        self.assertEqual(list_node[5], syntax_node.ValueNode("1.0", float))
        for val in list_node[1:5]:
            self.assertEqual(val, syntax_node.ValueNode("1.0", float))
        for val in list_node[1:5:1]:
            self.assertEqual(val, syntax_node.ValueNode("1.0", float))
        for val in list_node[::1]:
            self.assertEqual(val, syntax_node.ValueNode("1.0", float))
        for val in list_node[5:1:-1]:
            self.assertEqual(val, syntax_node.ValueNode("1.0", float))
        for val in list_node[::-1]:
            self.assertEqual(val, syntax_node.ValueNode("1.0", float))
        with self.assertRaises(IndexError):
            list_node[50]

    def test_list_equality(self):
        list_node1 = syntax_node.ListNode("list")
        for i in range(20):
            list_node1.append(syntax_node.ValueNode("1.0", float))
        with self.assertRaises(TypeError):
            list_node1 == "hi"
        list2 = [syntax_node.ValueNode("1.0", float)] * 19
        self.assertTrue(not list_node1 == list2)
        list2 = [syntax_node.ValueNode("1.0", float)] * 20
        self.assertTrue(list_node1 == list2)
        list2 = [syntax_node.ValueNode("1.0", float)] * 19 + [
            syntax_node.ValueNode("1.5", float)
        ]
        self.assertTrue(list_node1 != list2)

    def test_list_trailing_comment(self):
        list_node1 = syntax_node.ListNode("list")
        for i in range(20):
            list_node1.append(syntax_node.ValueNode("1.0", float))
        padding = syntax_node.PaddingNode("c hi", True)
        list_node1[-1].padding = padding
        comments = list(list_node1.get_trailing_comment())
        self.assertEqual(len(comments), 1)
        list_node1._delete_trailing_comment()
        self.assertIsNone(list_node1.get_trailing_comment())
        # test an empty list
        list_node1 = syntax_node.ListNode("list")
        self.assertIsNone(list_node1.get_trailing_comment())
        list_node1._delete_trailing_comment()
        self.assertIsNone(list_node1.get_trailing_comment())

    def test_list_format(self):
        list_node = syntax_node.ListNode("list")
        for i in range(20):
            list_node.append(syntax_node.ValueNode("1.0", float))
        self.assertEqual(list_node.format(), "1.0 " * 19 + "1.0")

    def test_list_comments(self):
        list_node = syntax_node.ListNode("list")
        for i in range(20):
            list_node.append(syntax_node.ValueNode("1.0", float))
        padding = syntax_node.PaddingNode("$ hi", True)
        list_node[-1].padding = padding
        comments = list(list_node.comments)
        self.assertEqual(len(comments), 1)


class TestIsotopesNode(TestCase):
    def test_isotopes_init(self):
        isotope = syntax_node.IsotopesNode("test")
        self.assertEqual(isotope.name, "test")
        self.assertIsInstance(isotope.nodes, list)

    def test_isotopes_append(self):
        isotopes = syntax_node.IsotopesNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        concentration = syntax_node.ValueNode("1.5", float)
        isotopes.append(("isotope_fraction", zaid, concentration))
        self.assertEqual(isotopes.nodes[-1][0], zaid)
        self.assertEqual(isotopes.nodes[-1][1], concentration)

    def test_isotopes_format(self):
        padding = syntax_node.PaddingNode(" ")
        isotopes = syntax_node.IsotopesNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        zaid.padding = padding
        concentration = syntax_node.ValueNode("1.5", float)
        concentration.padding = padding
        isotopes.append(("isotope_fraction", zaid, concentration))
        self.assertEqual(isotopes.format(), "1001.80c 1.5 ")

    def test_isotopes_str(self):
        isotopes = syntax_node.IsotopesNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        concentration = syntax_node.ValueNode("1.5", float)
        isotopes.append(("isotope_fraction", zaid, concentration))
        str(isotopes)
        repr(isotopes)

    def test_isotopes_iter(self):
        isotopes = syntax_node.IsotopesNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        concentration = syntax_node.ValueNode("1.5", float)
        isotopes.append(("isotope_fraction", zaid, concentration))
        isotopes.append(("isotope_fraction", zaid, concentration))
        for combo in isotopes:
            self.assertEqual(len(combo), 2)

    def test_isotopes_comments(self):
        padding = syntax_node.PaddingNode(" ")
        isotopes = syntax_node.IsotopesNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        zaid.padding = padding
        concentration = syntax_node.ValueNode("1.5", float)
        padding = copy.deepcopy(padding)
        padding.append("$ hi", True)
        concentration.padding = padding
        isotopes.append(("isotope_fraction", zaid, concentration))
        comments = list(isotopes.comments)
        self.assertEqual(len(comments), 1)
        self.assertEqual(comments[0].contents, "hi")

    def test_isotopes_trailing_comment(self):
        padding = syntax_node.PaddingNode(" ")
        isotopes = syntax_node.IsotopesNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        zaid.padding = padding
        concentration = syntax_node.ValueNode("1.5", float)
        padding = copy.deepcopy(padding)
        padding.append("c hi", True)
        concentration.padding = padding
        isotopes.append(("isotope_fraction", zaid, concentration))
        comments = isotopes.get_trailing_comment()
        self.assertEqual(len(comments), 1)
        self.assertEqual(comments[0].contents, "hi")
        isotopes._delete_trailing_comment()
        comments = isotopes.get_trailing_comment()
        self.assertIsNone(comments)


class TestShortcutNode(TestCase):
    def test_basic_shortcut_init(self):
        with self.assertRaises(ValueError):
            syntax_node.ShortcutNode("")
        # test a blank init
        shortcut = syntax_node.ShortcutNode(
            short_type=syntax_node.Shortcuts.LOG_INTERPOLATE
        )
        self.assertEqual(shortcut._type, syntax_node.Shortcuts.LOG_INTERPOLATE)
        self.assertEqual(shortcut.end_padding.nodes, [" "])
        with self.assertRaises(TypeError):
            syntax_node.ShortcutNode(short_type="")

    def test_shortcut_end_padding_setter(self):
        short = syntax_node.ShortcutNode()
        pad = syntax_node.PaddingNode(" ")
        short.end_padding = pad
        self.assertEqual(short.end_padding, pad)
        with self.assertRaises(TypeError):
            short.end_padding = " "


"""
Most examples, unless otherwise noted are taken from Section 2.8.1
of LA-UR-17-29981.
"""
tests = {
    "1 3M 2r": [1, 3, 3, 3],
    # unofficial
    "0.01 2ILOG 10": [0.01, 0.1, 1, 10],
    "1 3M I 4": [1, 3, 3.5, 4],
    "1 3M 3M": [1, 3, 9],
    "1 2R 2I 2.5": [1, 1, 1, 1.5, 2, 2.5],
    "1 R 2m": [1, 1, 2],
    "1 R R": [1, 1, 1],
    "1 2i 4 3m": [1, 2, 3, 4, 12],
    # unofficial
    "1 i 3": [1, 2, 3],
    # unofficial
    "1 ilog 100": [1, 10, 100],
    "1 1r ilog 100": [1, 1, 10, 100],
    # last official one
    "1 2i 4 2i 10": [
        1,
        2,
        3,
        4,
        6,
        8,
        10,
    ],
    "1 2j 4": [1, montepy.Jump(), montepy.Jump(), 4],
    "1 j": [1, montepy.Jump()],
}


@pytest.mark.parametrize("test, answer", tests.items())
def test_shortcut_expansion_valid(test, answer):
    parser = ShortcutTestFixture()
    print(test)
    input = Input([test], BlockType.DATA)
    parsed = parser.parse(input.tokenize())
    print(parsed)
    for val, gold in zip(parsed, answer):
        if val.value is None:
            assert gold == montepy.Jump()
        else:
            assert val.value == pytest.approx(gold)
    assert parsed.format() == test


@pytest.mark.parametrize(
    "test",
    [
        ("3J 4R"),
        ("1 4I 3M"),
        # last official test
        ("1 4I J"),
        ("1 2Ilog J"),
        ("J 2Ilog 5"),
        ("3J 2M"),
        ("10 M"),
        ("2R"),
    ],
)
def test_shortcut_expansion_invalid(test):
    print(test)
    parser = ShortcutTestFixture()
    with pytest.raises(ValueError):
        input = Input([test], BlockType.DATA)
        parsed = parser.parse(input.tokenize())
        if parsed is None:
            raise montepy.errors.MalformedInputError("", "")


@pytest.mark.parametrize(
    "test, answer, form_ans",
    [
        ("1 3r ", [1, 1, 1, 1], None),
        ("1 1 3r ", [1, 1, 1, 1, 1], None),
        ("1 1 2M 3r ", [1, 1, 2, 2, 2, 2], None),
        ("1 -2M ", [1, -2], None),
        ("1 2i 4 ", [1, 2, 3, 4], None),
        ("1 2i 4 ", [1, 2, 3, 4], None),
        ("1 1 2i 4 ", [1, 1, 2, 3, 4], None),
        ("1 1 2i 4 5 6 ", [1, 1, 2, 3, 4, 5, 6], "1 1 2I 4  5 6"),
        ("1 1 2i 4:5 6 ", [1, 1, 2, 3, 4, 5, 6], "1 1 2I 4 :5 6"),
        ("1 ilog 100 ", [1, 10, 100], "1 1ILOG 100"),
        ("1 1r ilog 100 ", [1, 1, 10, 100], "1 1R 1ILOG 100"),
        # secretly test iterator
        ("#1", [1], None),
        ("#(1 2 3)", [1, 2, 3], None),
        ("1 2:( 3 4 5)", [1, 2, 3, 4, 5], None),
    ],
)
def test_shortcut_geometry_expansion(test, answer, form_ans):

    parser = ShortcutGeometryTestFixture()
    print(test)
    input = Input([test], BlockType.CELL)
    parsed = parser.parse(input.tokenize())
    for val, gold in zip(parsed, answer):
        assert val.value == gold
    if form_ans:
        assert parsed.format().rstrip() == form_ans
    else:
        assert parsed.format().rstrip() == test.upper().rstrip()


@pytest.mark.parametrize(
    "test, length, indices",
    [
        ("1 3r ", 1, {0: Shortcuts.REPEAT}),
        ("1 1 3r ", 2, {1: Shortcuts.REPEAT}),
        ("1 1 2M 3r ", 3, {1: Shortcuts.MULTIPLY, 2: Shortcuts.REPEAT}),
        ("1 -2M ", 1, {0: Shortcuts.MULTIPLY}),
        ("1 2i 4 ", 1, {0: Shortcuts.INTERPOLATE}),
        ("1 2i 4 ", 1, {0: Shortcuts.INTERPOLATE}),
        ("1 1 2i 4 ", 2, {1: Shortcuts.INTERPOLATE}),
        ("1 1 2i 4 5 6 ", 4, {1: Shortcuts.INTERPOLATE}),
        ("1 1 2i 4:5 6 ", 4, {1: Shortcuts.INTERPOLATE}),
        ("1 ilog 100 ", 1, {0: Shortcuts.LOG_INTERPOLATE}),
        # secretly test iterator
        ("#1", 1, {}),
        ("#(1 2 3)", 3, {}),
        ("1 2:( 3 4 5)", 5, {}),
    ],
)
def test_shortcut_flatten(test, length, indices):

    parser = ShortcutGeometryTestFixture()
    print(test)
    input = Input([test], BlockType.CELL)
    parsed = parser.parse(input.tokenize())
    flatpack = parsed._flatten_shortcut()
    assert len(flatpack) == length
    for index, short_type in indices.items():
        assert isinstance(flatpack.nodes[index], syntax_node.ShortcutNode)
        assert flatpack.nodes[index].type == short_type


@pytest.mark.parametrize(
    "in_str, answer",
    [
        ("1 5R", "1 5R"),
        ("1 5r", "1 5r"),
        ("1 5r 2m", "1 5r 2m"),
        ("1 r", "1 r"),
        ("1 r 2 2r", "1 r 2 2r"),
        ("1 J 5 2R", "1 J 5 2R"),
        ("1 2i 5", "1 2i 5"),
        ("1 2ilog 5", "1 2ilog 5"),
        ("1 r 2i 5", "1 r 2i 5"),
        ("1 2i 5 r", "1 2i 5 r"),
        ("1 r 2ilog 5", "1 r 2ilog 5"),
        ("1 5M", "1 5M"),
        ("1 5m", "1 5m"),
        ("1 5m 2r", "1 5m 2r"),
        ("2j ", "2j "),
        ("2j", "2j"),
        ("2J ", "2J "),
        ("J", "J"),
        ("0 2i 3 10 2i 16 1", "0 2i 3 10 2i 16 1"),
    ],
)
def test_shortcut_format(in_str, answer):
    parser = ShortcutTestFixture()
    input = Input([in_str], BlockType.CELL)
    shortcut = parser.parse(input.tokenize())
    assert shortcut.format() == answer
    # try jump with empty jump shortcut
    shortcut.nodes.clear()
    assert shortcut.format() == ""


class ShortcutTestFixture(MCNP_Parser):
    @_("number_sequence", "shortcut_sequence")
    def shortcut_magic(self, p):
        return p[0]


class ShortcutGeometryTestFixture(montepy.input_parser.cell_parser.CellParser):
    @_("geometry_expr")
    def geometry(self, p):
        return p[0]


class TestShortcutListIntegration(TestCase):
    def setUp(self):
        self.parser = ShortcutTestFixture()
        input = Input(["1 1 2i 4"], BlockType.DATA)
        self.list_node = self.parser.parse(input.tokenize())

    def test_shortcut_list_node_init(self):
        answers = [1, 1, 2, 3, 4]
        for val, gold in zip(self.list_node, answers):
            self.assertAlmostEqual(val.value, gold)

    def test_shortcut_list_update_vals(self):
        list_node = copy.deepcopy(self.list_node)
        values = list(list_node)
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)

    def test_shortcut_list_update_vals_repeat(self):
        input = Input(["1 2 3 5R 0 0"], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.insert(2, syntax_node.ValueNode(3.0, float))
        list_node.update_with_new_values(values)
        self.assertEqual(len(list_node._shortcuts[0].nodes), 7)
        self.assertEqual(list(list_node), values)

    def test_shortcut_list_trailing_jump(self):
        input = Input(["1 2 3 5R 0 0"], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode(Jump(), float))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values[:-1])
        # test with User specified end jump
        input = Input(["1 2 3 5R 0 0 j"], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode(Jump(), float))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)

    def test_shortcut_list_touching_shortcuts(self):
        input = Input(["1 2 3 5R 3 3 4R 0 0"], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)

    def test_shortcut_then_jump_compress(self):
        for input_str, index in [
            ("1 2 3 5R 3 3 4R 0 0", -2),
            ("1 2 3M 4", -2),
            ("1 2ilog 10", -2),
        ]:
            print(input_str)
            input = Input([input_str], BlockType.DATA)
            list_node = self.parser.parse(input.tokenize())
            values = list(list_node)
            values[index].value = None
            list_node.update_with_new_values(values)
            self.assertEqual(list(list_node), values)

    def test_shortcut_list_shortcut_cant_consume(self):
        # try with wrong type
        input = Input(["1 2 3 5R 3 3 4R "], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("hi", str))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)
        self.assertEqual(len(list_node._shortcuts[1].nodes), 5)
        # try with wrong value
        input = Input(["1 2 3 5R 3 3 4R "], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("5.0", float))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)
        self.assertEqual(len(list_node._shortcuts[1].nodes), 5)
        # try with right value
        values[-1].value = 3.0
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)
        self.assertEqual(len(list_node._shortcuts[1].nodes), 6)

    def test_shortcut_list_multiply(self):
        input = Input(["1 2 5M "], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("5.0", float))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)
        self.assertEqual(len(list_node._shortcuts[0].nodes), 2)
        input = Input(["0.5 2R 5M "], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("5.0", float))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)
        self.assertEqual(len(list_node._shortcuts[1].nodes), 1)

    def test_shortcut_list_interpolate(self):
        # try with log interpolate
        input = Input(["1.0 0.01 2ILOG 10"], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("100", float))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)
        self.assertEqual(len(list_node._shortcuts[0].nodes), 5)
        # try with linear interpolate
        input = Input(["1 1 2I 2.5"], BlockType.DATA)
        list_node = self.parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("3", float))
        list_node.update_with_new_values(values)
        self.assertEqual(list(list_node), values)
        self.assertEqual(len(list_node._shortcuts[0].nodes), 5)


class TestSyntaxParsing(TestCase):
    def testCardInit(self):
        with self.assertRaises(TypeError):
            Input("5", BlockType.CELL)
        with self.assertRaises(TypeError):
            Input([5], BlockType.CELL)
        with self.assertRaises(TypeError):
            Input(["5"], "5")

    def testMessageInit(self):
        with self.assertRaises(TypeError):
            Message(["hi"], "5")
        with self.assertRaises(TypeError):
            Message(["hi"], [5])

    def testTitleInit(self):
        with self.assertRaises(TypeError):
            Title(["hi"], 5)

    def testMessageFinder(self):
        test_message = "this is a message"
        test_string = f"""message: {test_message}

test title
"""
        for tester, validator in [
            (test_string, test_message),
            (test_string.upper(), test_message.upper()),
        ]:
            with StringIO(tester) as fh:
                generator = input_syntax_reader.read_front_matters(fh, (6, 2, 0))
                card = next(generator)
                self.assertIsInstance(card, montepy.input_parser.mcnp_input.Message)
                self.assertEqual(card.lines[0], validator)
                self.assertEqual(len(card.lines), 1)

    def testReadCardStr(self):
        card = ReadInput(["Read file=hi.imcnp"], BlockType.CELL)
        self.assertEqual(str(card), "READ INPUT: Block_Type: BlockType.CELL")
        self.assertEqual(
            repr(card),
            "READ INPUT: BlockType.CELL: ['Read file=hi.imcnp'] File: hi.imcnp",
        )

    def testSingleLineEndComment(self):
        # tests issues #117
        input = Input(["c"], BlockType.CELL)
        generator = input.tokenize()
        token = next(generator)
        self.assertEqual(token.type, "COMMENT")
        self.assertEqual(token.value, "c")

    def testReadCardConfusions(self):
        for file in {"A1_cells", "I1_cells"}:
            input = ReadInput([f"Read FILE={file}"], BlockType.CELL)
            self.assertEqual(input.file_name, file)

    def testReadCardBadSyntax(self):
        with self.assertRaises(ParsingError):
            card = ReadInput(["Read 1"], BlockType.CELL)

    def testTitleFinder(self):
        test_title = "Richard Stallman writes GNU"
        test_string = f"""{test_title}
1 0 -1
"""
        for tester, validator in [
            (test_string, test_title),
            (test_string.upper(), test_title.upper()),
        ]:
            with StringIO(tester) as fh:
                generator = input_syntax_reader.read_front_matters(fh, (6, 2, 0))
                card = next(generator)
                self.assertIsInstance(card, montepy.input_parser.mcnp_input.Title)
                self.assertEqual(card.title, validator)

    def testCardFinder(self):
        test_string = """1 0 -1
     5"""
        for i in range(5):
            tester = " " * i + test_string
            with StringIO(tester) as fh:
                fh.lineno = 0
                fh.path = "foo"
                generator = input_syntax_reader.read_data(fh, (6, 2, 0))
                card = next(generator)
                self.assertIsInstance(card, montepy.input_parser.mcnp_input.Input)
                answer = [" " * i + "1 0 -1", "     5"]
                self.assertEqual(len(answer), len(card.input_lines))
                for j, line in enumerate(card.input_lines):
                    self.assertEqual(line, answer[j])
                self.assertEqual(
                    card.block_type, montepy.input_parser.block_type.BlockType.CELL
                )

    # TODO ensure this is tested in Input parsers
    """
    def testCommentFinder(self):
        for i in range(5):
            tester = " " * i + test_string
            with StringIO(tester) as fh:
                card = next(input_syntax_reader.read_data(fh, (6, 2, 0)))
                self.assertIsInstance(card, montepy.input_parser.mcnp_input.Comment)
                self.assertEqual(len(card.lines), 5)
                self.assertEqual(card.lines[0], "foo")
                self.assertEqual(card.lines[1], "bar")
                self.assertEqual(card.lines[3], "bop")
    """

    def testReadCardFinder(self):
        test_string = "read file=foo.imcnp "
        with StringIO(test_string) as fh:
            fh.lineno = 0
            fh.path = "foo"
            card = next(input_syntax_reader.read_data(fh, (6, 2, 0)))
            self.assertIsNone(card)  # the read input is hidden from the user

    def testBlockId(self):
        test_string = "1 0 -1"
        for i in range(3):
            tester = "\n" * i + test_string
            with StringIO(tester) as fh:
                fh.lineno = 0
                fh.path = "foo"
                fh.name = "name"
                for card in input_syntax_reader.read_data(fh, (6, 2, 0)):
                    pass
                self.assertEqual(
                    montepy.input_parser.block_type.BlockType(i), card.block_type
                )

    def testCommentFormatInput(self):
        in_strs = ["c foo", "c bar"]
        card = montepy.input_parser.syntax_node.CommentNode(in_strs[0])
        output = card.format()
        answer = "c foo"
        str_answer = "COMMENT: c foo"
        self.assertEqual(repr(card), str_answer)
        self.assertEqual("c foo", str(card))
        self.assertEqual(len(answer), len(output))
        for i, line in enumerate(output):
            self.assertEqual(answer[i], line)

    def testMessageFormatInput(self):
        answer = ["MESSAGE: foo", "bar", ""]
        card = montepy.input_parser.mcnp_input.Message(answer, ["foo", "bar"])
        str_answer = """MESSAGE:
foo
bar
"""
        self.assertEqual(str_answer, repr(card))
        self.assertEqual("MESSAGE: 2 lines", str(card))
        output = card.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(len(answer), len(output))
        for i, line in enumerate(output):
            self.assertEqual(answer[i], line)

    def testTitleFormatInput(self):
        card = montepy.input_parser.mcnp_input.Title(["foo"], "foo")
        answer = ["foo"]
        str_answer = "TITLE: foo"
        self.assertEqual(str(card), str_answer)
        output = card.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(len(answer), len(output))
        for i, line in enumerate(output):
            self.assertEqual(answer[i], line)

    def testReadInput(self):
        # TODO ensure comments are properly glued to right input
        generator = input_syntax_reader.read_input_syntax(
            MCNP_InputFile("tests/inputs/test.imcnp")
        )
        mcnp_in = montepy.input_parser.mcnp_input
        input_order = [mcnp_in.Message, mcnp_in.Title]
        input_order += [mcnp_in.Input] * 17
        for i, input in enumerate(generator):
            print(input.input_lines)
            print(input_order[i])
            self.assertIsInstance(input, input_order[i])

    def testReadInputWithRead(self):
        generator = input_syntax_reader.read_input_syntax(
            MCNP_InputFile("tests/inputs/testRead.imcnp")
        )
        next(generator)  # skip title
        next(generator)  # skip read none
        next(generator)  # skip surfaces input
        next(generator)  # skip data mode input
        card = next(generator)
        answer = ["1 0 -1", "c"]
        self.assertEqual(answer, card.input_lines)

    def testReadInputWithVertMode(self):
        generator = input_syntax_reader.read_input_syntax(
            MCNP_InputFile("tests/inputs/testVerticalMode.imcnp")
        )
        next(generator)
        next(generator)
        with self.assertRaises(montepy.errors.UnsupportedFeature):
            next(generator)

    def testCardStringRepr(self):
        in_str = "1 0 -1"
        card = montepy.input_parser.mcnp_input.Input(
            [in_str], montepy.input_parser.block_type.BlockType.CELL
        )
        self.assertEqual(str(card), "INPUT: BlockType.CELL")
        self.assertEqual(repr(card), "INPUT: BlockType.CELL: ['1 0 -1']")

    def testDataInputNameParsing(self):
        tests = {
            "kcOde": {"prefix": "kcode", "number": None, "classifier": None},
            "M300": {"prefix": "m", "number": 300, "classifier": None},
            "IMP:N,P,E": {
                "prefix": "imp",
                "number": None,
                "classifier": [Particle.NEUTRON, Particle.PHOTON, Particle.ELECTRON],
            },
            "F1004:n,P": {
                "prefix": "f",
                "number": 1004,
                "classifier": [Particle.NEUTRON, Particle.PHOTON],
            },
        }
        for in_str, answer in tests.items():
            # Testing parsing the names
            print("in", in_str, "answer", answer)
            card = montepy.input_parser.mcnp_input.Input(
                [in_str], montepy.input_parser.block_type.BlockType.DATA
            )
            data_input = montepy.data_inputs.data_input.DataInput(card, fast_parse=True)
            self.assertEqual(data_input.prefix, answer["prefix"])
            if answer["number"]:
                self.assertEqual(data_input._input_number.value, answer["number"])
            if answer["classifier"]:
                self.assertEqual(
                    sorted(data_input.particle_classifiers),
                    sorted(answer["classifier"]),
                )

    def testDataInputNameEnforcement(self):
        tests = {
            "kcOde5": {"prefix": "kcode", "number": False, "classifier": 0},
            "M-300": {"prefix": "m", "number": True, "classifier": 0},
            "M": {"prefix": "m", "number": True, "classifier": 0},
            "f4m": {"prefix": "fm", "number": True, "classifier": 1},
            "IMP:N,P,E": {"prefix": "imp", "number": False, "classifier": 0},
            "IMP": {"prefix": "imp", "number": False, "classifier": 2},
        }
        valid = {
            "IMP:N,P,E": {"prefix": "imp", "number": False, "classifier": 2},
            "F1004:n,P": {"prefix": "f", "number": True, "classifier": 1},
        }
        # tests invalid names
        for in_str, answer in tests.items():
            with self.assertRaises(montepy.errors.MalformedInputError):
                card = montepy.input_parser.mcnp_input.Input(
                    [in_str], montepy.input_parser.block_type.BlockType.DATA
                )
                Fixture = DataInputTestFixture
                Fixture._class_prefix1 = answer["prefix"]
                Fixture._has_number1 = answer["number"]
                Fixture._has_classifier1 = answer["classifier"]
                card = Fixture(card)

        # tests valid names
        for in_str, answer in valid.items():
            card = montepy.input_parser.mcnp_input.Input(
                [in_str], montepy.input_parser.block_type.BlockType.DATA
            )
            print(card.input_lines)
            print(
                "Prefix",
                answer["prefix"],
                "number",
                answer["number"],
                "classifier",
                answer["classifier"],
            )
            Fixture = DataInputTestFixture
            Fixture._class_prefix1 = answer["prefix"]
            Fixture._has_number1 = answer["number"]
            Fixture._has_classifier1 = answer["classifier"]
            card = Fixture(card)

    def test_get_line_numbers(self):
        answers = {
            (5, 1, 60): 80,
            (6, 1, 0): 80,
            (6, 2, 0): 128,
            (6, 2, 3): 128,
            (6, 3, 0): 128,
            (7, 4, 0): 128,
        }
        for version, answer in answers.items():
            self.assertEqual(answer, montepy.constants.get_max_line_length(version))
        with self.assertRaises(montepy.errors.UnsupportedFeature):
            montepy.constants.get_max_line_length((5, 1, 38))

    def test_jump(self):
        jump = Jump()
        self.assertEqual("J", str(jump))
        jump2 = Jump()
        self.assertEqual(jump, jump2)
        with self.assertRaises(TypeError):
            bool(jump)

    def test_jump_and_a_hop(self):
        jump = Jump()
        # first you need to hop
        self.assertEqual("j", jump.lower())
        # then you need to skip
        self.assertEqual("Jump", jump.title())
        # before you can jump
        self.assertEqual("J", jump.upper())
        str(jump)
        repr(jump)


class TestClassifierNode(TestCase):
    def test_classifier_init(self):
        classifier = syntax_node.ClassifierNode()
        self.assertIsNone(classifier.prefix)
        self.assertIsNone(classifier.number)
        self.assertIsNone(classifier.particles)
        self.assertIsNone(classifier.modifier)
        self.assertIsNone(classifier.padding)

    def test_classifier_setter(self):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("M", str)
        self.assertEqual(classifier.prefix.value, "M")
        classifier.number = syntax_node.ValueNode("124", int)
        self.assertEqual(classifier.number.value, 124)
        classifier.modifier = syntax_node.ValueNode("*", str)
        self.assertEqual(classifier.modifier.value, "*")
        classifier.padding = syntax_node.PaddingNode(" ")
        self.assertEqual(len(classifier.padding.nodes), 1)

    def test_classifier_format(self):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("M", str)
        self.assertEqual(classifier.format(), "M")
        classifier.number = syntax_node.ValueNode("124", int)
        self.assertEqual(classifier.format(), "M124")
        classifier.modifier = syntax_node.ValueNode("*", str)
        self.assertEqual(classifier.format(), "*M124")
        classifier.padding = syntax_node.PaddingNode(" ")
        self.assertEqual(classifier.format(), "*M124 ")
        str(classifier)
        repr(classifier)

    def test_classifier_comments(self):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("M", str)
        self.assertEqual(len(list(classifier.comments)), 0)
        classifier.padding = syntax_node.PaddingNode(" ")
        classifier.padding.append("$ hi", True)
        self.assertEqual(len(list(classifier.comments)), 1)


class TestParametersNode(TestCase):
    def setUp(self):
        self.param = syntax_node.ParametersNode()
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("vol", str)
        list_node = syntax_node.ListNode("data")
        list_node.append(syntax_node.ValueNode("1.0", float))
        self.param.append(
            syntax_node.SyntaxNode(
                "hi",
                {
                    "classifier": classifier,
                    "seperator": syntax_node.ValueNode("=", str),
                    "data": list_node,
                },
            )
        )

    def test_parameter_init(self):
        param = syntax_node.ParametersNode()
        self.assertIsInstance(param.nodes, dict)

    def test_parameter_append(self):
        self.assertEqual(len(self.param.nodes), 1)
        with self.assertRaises(ValueError):
            classifier = syntax_node.ClassifierNode()
            classifier.prefix = syntax_node.ValueNode("vol", str)
            self.param.append(syntax_node.SyntaxNode("foo", {"classifier": classifier}))

    def test_parameter_dict(self):
        param = self.param
        self.assertEqual(param["vol"], param.nodes["vol"])
        with self.assertRaises(KeyError):
            param["hi"]
        self.assertIn("vol", param)

    def test_parameter_str(self):
        str(self.param)
        repr(self.param)

    def test_parameter_format(self):
        self.assertEqual(self.param.format(), "vol=1.0")

    def test_parameter_comments(self):
        param = copy.deepcopy(self.param)
        self.assertEqual(len(list(param.comments)), 0)
        param["vol"]["data"][0].padding = syntax_node.PaddingNode(" ")
        param["vol"]["data"][0].padding.append("$ hi", True)
        self.assertEqual(len(list(param.comments)), 1)

    def test_parameter_trailing_comments(self):
        param = copy.deepcopy(self.param)
        self.assertIsNone(param.get_trailing_comment())
        param._delete_trailing_comment()
        self.assertIsNone(param.get_trailing_comment())
        padding = syntax_node.PaddingNode("$ hi", True)
        param["vol"]["data"][0].padding = padding
        comment = param.get_trailing_comment()
        self.assertIsNone(comment)
        padding.append(syntax_node.CommentNode("c hi"), True)
        comment = param.get_trailing_comment()
        self.assertEqual(comment[0].contents, "hi")
        param._delete_trailing_comment()
        self.assertIsNone(param.get_trailing_comment())


class DataInputTestFixture(montepy.data_inputs.data_input.DataInputAbstract):
    _class_prefix1 = None
    _has_number1 = None
    _has_classifier1 = None

    def __init__(self, input_card=None):
        """
        :param input_card: the Card object representing this data input
        :type input_card: Input
        """
        super().__init__(input_card, fast_parse=True)

    def _class_prefix(self):
        return self._class_prefix1

    def _has_number(self):
        return self._has_number1

    def _has_classifier(self):
        return self._has_classifier1



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
import io
from pathlib import Path

import pytest
import os

import montepy
from montepy.data_inputs import material, volume
from montepy.input_parser.mcnp_input import (
    Input,
    Jump,
    Message,
    Title,
)
from montepy.errors import *
from montepy.particle import Particle
import numpy as np

from tests import constants


@pytest.fixture(scope="module")
def simple_problem():
    return montepy.read_input(os.path.join("tests", "inputs", "test.imcnp"))


@pytest.fixture(scope="module")
def importance_problem():
    return montepy.read_input(os.path.join("tests", "inputs", "test_importance.imcnp"))


@pytest.fixture(scope="module")
def universe_problem():
    return montepy.read_input(os.path.join("tests", "inputs", "test_universe.imcnp"))


@pytest.fixture(scope="module")
def data_universe_problem():
    return montepy.read_input(
        os.path.join("tests", "inputs", "test_universe_data.imcnp")
    )


def test_original_input(simple_problem):
    cell_order = [Message, Title] + [Input] * 17
    for i, input_ob in enumerate(simple_problem.original_inputs):
        assert isinstance(input_ob, cell_order[i])


def test_original_input_dos():
    dos_problem = montepy.read_input(os.path.join("tests", "inputs", "test_dos.imcnp"))
    cell_order = [Message, Title] + [Input] * 16
    for i, input_ob in enumerate(dos_problem.original_inputs):
        assert isinstance(input_ob, cell_order[i])


def test_original_input_tabs():
    problem = montepy.read_input(os.path.join("tests", "inputs", "test_tab.imcnp"))
    cell_order = [Message, Title] + [Input] * 17
    for i, input_ob in enumerate(problem.original_inputs):
        assert isinstance(input_ob, cell_order[i])


# TODO formalize this or see if this is covered by other tests.
def test_lazy_comments_check(simple_problem):
    material2 = simple_problem.materials[2]
    for comment in material2._tree.comments:
        print(repr(comment))
    print(material2._tree.get_trailing_comment())


def test_moving_trail_comments(universe_problem):
    problem = copy.deepcopy(universe_problem)
    mat = problem.materials[2]
    idx = problem.data_inputs.index(mat)
    dat = problem.data_inputs[idx - 1]
    assert len(mat.comments) == 1
    assert len(mat.leading_comments) == 1
    assert len(dat.leading_comments) == 0
    assert len(dat.comments) == 0


def test_material_parsing(simple_problem):
    mat_numbers = [1, 2, 3]
    for i, mat in enumerate(simple_problem.materials):
        assert mat.number == mat_numbers[i]


def test_surface_parsing(simple_problem):
    surf_numbers = [1000, 1005, 1010]
    for i, surf in enumerate(simple_problem.surfaces):
        assert surf.number == surf_numbers[i]


def test_data_card_parsing(simple_problem):
    M = material.Material
    V = volume.Volume
    cards = [M, M, M, "KSRC", "KCODE", "PHYS:P", "MODE", V]
    for i, card in enumerate(simple_problem.data_inputs):
        if isinstance(cards[i], str):
            assert card.classifier.format().upper().rstrip() == cards[i]
        else:
            assert isinstance(card, cards[i])
        if i == 2:
            assert card.thermal_scattering is not None


def test_cells_parsing_linking(simple_problem):
    cell_numbers = [1, 2, 3, 99, 5]
    mats = simple_problem.materials
    mat_answer = [mats[1], mats[2], mats[3], None, None]
    surfs = simple_problem.surfaces
    surf_answer = [{surfs[1000]}, {surfs[1005]}, set(surfs), {surfs[1010]}, set()]
    cells = simple_problem.cells
    complements = [set()] * 4 + [{cells[99]}]
    for i, cell in enumerate(simple_problem.cells):
        assert cell.number == cell_numbers[i]
        assert cell.material == mat_answer[i]
        surfaces = set(cell.surfaces)
        assert surfaces.union(surf_answer[i]) == surfaces
        assert set(cell.complements).union(complements[i]) == complements[i]


def test_message(simple_problem):
    lines = ["this is a message", "it should show up at the beginning", "foo"]
    for i, line in enumerate(simple_problem.message.lines):
        assert line == lines[i]


def test_title(simple_problem):
    answer = "MCNP Test Model for MOAA"
    assert answer == simple_problem.title.title


def test_read_card_recursion():
    problem = montepy.read_input("tests/inputs/testReadRec1.imcnp")
    assert len(problem.cells) == 1
    assert len(problem.surfaces) == 1
    assert montepy.particle.Particle.PHOTON in problem.mode


def test_problem_str(simple_problem):
    output = str(simple_problem)
    assert "MCNP problem for: tests/inputs/test.imcnp" in output


def test_write_to_file(simple_problem):
    out = "foo.imcnp"
    try:
        problem = copy.deepcopy(simple_problem)
        problem.write_to_file(out)
        with open(out, "r") as fh:
            for line in fh:
                print(line.rstrip())
        test_problem = montepy.read_input(out)
        for i, cell in enumerate(simple_problem.cells):
            num = cell.number
            assert num == test_problem.cells[num].number
            for attr in {"universe", "volume"}:
                print(f"testing the attribute: {attr}")
                gold = getattr(simple_problem.cells[num], attr)
                test = getattr(test_problem.cells[num], attr)
                if attr in {"universe"} and gold is not None:
                    gold, test = (gold.number, test.number)
                assert gold == test
        for i, surf in enumerate(simple_problem.surfaces):
            num = surf.number
            assert surf.number == test_problem.surfaces[num].number
        for i, data in enumerate(simple_problem.data_inputs):
            if isinstance(data, material.Material):
                assert data.number == test_problem.data_inputs[i].number
                if data.thermal_scattering is not None:
                    assert test_problem.data_inputs[i].thermal_scattering is not None
            elif isinstance(data, volume.Volume):
                assert str(data) == str(test_problem.data_inputs[i])
            else:
                print("Rewritten data", data.data)
                print("Original input data", test_problem.data_inputs[i].data)
                assert data.data == test_problem.data_inputs[i].data
    finally:
        if os.path.exists(out):
            os.remove(out)


def test_cell_material_setter(simple_problem):
    cell = copy.deepcopy(simple_problem.cells[1])
    mat = simple_problem.materials[2]
    cell.material = mat
    assert cell.material == mat
    cell.material = None
    assert cell.material is None
    with pytest.raises(TypeError):
        cell.material = 5


def test_problem_cells_setter(simple_problem):
    problem = copy.deepcopy(simple_problem)
    cells = copy.deepcopy(simple_problem.cells)
    cells.remove(cells[1])
    with pytest.raises(TypeError):
        problem.cells = 5
    with pytest.raises(TypeError):
        problem.cells = [5]
    with pytest.raises(TypeError):
        problem.cells.append(5)
    problem.cells = cells
    assert problem.cells.objects == cells.objects
    problem.cells = list(cells)
    assert problem.cells[2] == cells[2]
    # test that cell modifiers are still there
    problem.cells._importance.format_for_mcnp_input((6, 2, 0))


def test_problem_test_setter(simple_problem):
    problem = copy.deepcopy(simple_problem)
    sample_title = "This is a title"
    problem.title = sample_title
    assert problem.title.title == sample_title
    with pytest.raises(TypeError):
        problem.title = 5


def test_problem_children_adder(simple_problem):
    problem = copy.deepcopy(simple_problem)
    BT = montepy.input_parser.block_type.BlockType
    in_str = "5 SO 5.0"
    card = montepy.input_parser.mcnp_input.Input([in_str], BT.SURFACE)
    surf = montepy.surfaces.surface_builder.surface_builder(card)
    in_str = "M5 6000.70c 1.0"
    card = montepy.input_parser.mcnp_input.Input([in_str], BT.SURFACE)
    mat = montepy.data_inputs.material.Material(card)
    in_str = "TR1 0 0 1"
    input = montepy.input_parser.mcnp_input.Input([in_str], BT.DATA)
    transform = montepy.data_inputs.transform.Transform(input)
    surf.transform = transform
    cell_num = 1000
    cell = montepy.Cell()
    cell.material = mat
    cell.geometry = -surf
    cell.mass_density = 1.0
    cell.number = cell_num
    cell.universe = problem.universes[350]
    problem.cells.append(cell)
    problem.add_cell_children_to_problem()
    assert surf in problem.surfaces
    assert mat in problem.materials
    assert mat in problem.data_inputs
    assert transform in problem.transforms
    assert transform in problem.data_inputs
    for cell_num in [1, cell_num]:
        print(cell_num)
        if cell_num == 1000:
            with pytest.warns(LineExpansionWarning):
                output = problem.cells[cell_num].format_for_mcnp_input((6, 2, 0))
        else:
            output = problem.cells[cell_num].format_for_mcnp_input((6, 2, 0))
        print(output)
        assert "U=350" in "\n".join(output).upper()


def test_problem_mcnp_version_setter(simple_problem):
    problem = copy.deepcopy(simple_problem)
    with pytest.raises(ValueError):
        problem.mcnp_version = (4, 5, 3)
    problem.mcnp_version = (6, 2, 5)
    assert problem.mcnp_version == (6, 2, 5)


def test_problem_duplicate_surface_remover():
    problem = montepy.read_input("tests/inputs/test_redundant_surf.imcnp")
    nums = list(problem.surfaces.numbers)
    survivors = (
        nums[0:3] + nums[5:8] + [nums[9]] + nums[11:13] + [nums[13]] + [nums[-2]]
    )
    problem.remove_duplicate_surfaces(1e-4)
    assert list(problem.surfaces.numbers) == survivors
    cell_surf_answer = "-1 3 -6"
    assert cell_surf_answer in problem.cells[2].format_for_mcnp_input((6, 2, 0))[1]


def test_surface_periodic():
    problem = montepy.read_input("tests/inputs/test_surfaces.imcnp")
    surf = problem.surfaces[1]
    periodic = problem.surfaces[2]
    assert surf.periodic_surface == periodic
    assert "1 -2 SO" in surf.format_for_mcnp_input((6, 2, 0))[0]
    surf.periodic_surface = problem.surfaces[3]
    assert surf.periodic_surface == problem.surfaces[3]
    del surf.periodic_surface
    assert surf.periodic_surface is None
    with pytest.raises(TypeError):
        surf.periodic_surface = 5


def test_surface_transform():
    problem = montepy.read_input("tests/inputs/test_surfaces.imcnp")
    surf = problem.surfaces[1]
    transform = problem.data_inputs[0]
    del surf.periodic_surface
    surf.transform = transform
    assert surf.transform == transform
    assert "1 1 SO" in surf.format_for_mcnp_input((6, 2, 0))[0]
    del surf.transform
    assert surf.transform is None
    assert "1 SO" in surf.format_for_mcnp_input((6, 2, 0))[0]
    with pytest.raises(TypeError):
        surf.transform = 5


def test_materials_setter(simple_problem):
    problem = copy.deepcopy(simple_problem)
    with pytest.raises(TypeError):
        problem.materials = 5
    with pytest.raises(TypeError):
        problem.materials = [5]
    size = len(problem.materials)
    problem.materials = list(problem.materials)
    assert len(problem.materials) == size
    problem.materials = problem.materials
    assert len(problem.materials) == size


def test_reverse_pointers(simple_problem):
    problem = simple_problem
    complements = list(problem.cells[99].cells_complementing_this)
    assert problem.cells[5] in complements
    assert len(complements) == 1
    cells = list(problem.materials[1].cells)
    assert problem.cells[1] in cells
    assert len(cells) == 1
    cells = list(problem.surfaces[1005].cells)
    assert problem.cells[2] in problem.surfaces[1005].cells
    assert len(cells) == 2


def test_surface_card_pass_through():
    problem = montepy.read_input("tests/inputs/test_surfaces.imcnp")
    surf = problem.surfaces[1]
    # Test input pass through
    answer = ["1 -2 SO -5"]
    assert surf.format_for_mcnp_input((6, 2, 0)) == answer
    # Test changing periodic surface
    new_prob = copy.deepcopy(problem)
    surf = new_prob.surfaces[1]
    new_prob.surfaces[2].number = 5
    assert int(surf.format_for_mcnp_input((6, 2, 0))[0].split()[1]) == -5
    # Test changing transform
    new_prob = copy.deepcopy(problem)
    surf = new_prob.surfaces[4]
    surf.transform.number = 5
    assert int(surf.format_for_mcnp_input((6, 2, 0))[0].split()[1]) == 5
    # test changing surface constants
    new_prob = copy.deepcopy(problem)
    surf = new_prob.surfaces[4]
    surf.location = 2.5
    assert float(surf.format_for_mcnp_input((6, 2, 0))[0].split()[-1]) == 2.5


def test_surface_broken_link():
    with pytest.raises(montepy.errors.MalformedInputError):
        montepy.read_input("tests/inputs/test_broken_surf_link.imcnp")
    with pytest.raises(montepy.errors.MalformedInputError):
        montepy.read_input("tests/inputs/test_broken_transform_link.imcnp")


def test_material_broken_link():
    with pytest.raises(montepy.errors.BrokenObjectLinkError):
        problem = montepy.read_input("tests/inputs/test_broken_mat_link.imcnp")


def test_cell_surf_broken_link():
    with pytest.raises(montepy.errors.BrokenObjectLinkError):
        problem = montepy.read_input("tests/inputs/test_broken_cell_surf_link.imcnp")


def test_cell_complement_broken_link():
    with pytest.raises(montepy.errors.BrokenObjectLinkError):
        problem = montepy.read_input("tests/inputs/test_broken_complement.imcnp")


def test_cell_card_pass_through(simple_problem):
    problem = copy.deepcopy(simple_problem)
    cell = problem.cells[1]
    # test input pass-through
    answer = [
        "C cells",
        "c # hidden vertical Do not touch",
        "c",
        "1 1 20",
        "         -1000  $ dollar comment",
        "        imp:n,p=1 U=350 trcl=5 ",
    ]
    assert cell.format_for_mcnp_input((6, 2, 0)) == answer
    # test surface change
    new_prob = copy.deepcopy(problem)
    new_prob.surfaces[1000].number = 5
    cell = new_prob.cells[1]
    output = cell.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert int(output[4].split("$")[0]) == -5
    # test mass density printer
    cell.mass_density = 10.0
    with pytest.warns(LineExpansionWarning):
        output = cell.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert pytest.approx(float(output[3].split()[2])) == -10
    # ensure that surface number updated
    # Test material number change
    new_prob = copy.deepcopy(problem)
    new_prob.materials[1].number = 5
    cell = new_prob.cells[1]
    output = cell.format_for_mcnp_input((6, 2, 0))
    assert int(output[3].split()[1]) == 5


def test_thermal_scattering_pass_through(simple_problem):
    problem = copy.deepcopy(simple_problem)
    mat = problem.materials[3]
    therm = mat.thermal_scattering
    mat.number = 5
    assert therm.format_for_mcnp_input((6, 2, 0)) == ["MT5 lwtr.23t h-zr.20t h/zr.28t"]


def test_cutting_comments_parse():
    problem = montepy.read_input("tests/inputs/breaking_comments.imcnp")
    comments = problem.cells[1].comments
    assert len(comments) == 3
    assert "this is a cutting comment" in list(comments)[2].contents
    comments = problem.materials[2].comments
    assert len(comments) == 2


def test_cutting_comments_print_no_mutate():
    problem = montepy.read_input("tests/inputs/breaking_comments.imcnp")
    cell = problem.cells[1]
    output = cell.format_for_mcnp_input((6, 2, 0))
    assert len(output) == 5
    assert "c this is a cutting comment" == output[3]
    material2 = problem.materials[2]
    output = material2.format_for_mcnp_input((6, 2, 0))
    assert len(output) == 5
    assert "c          26057.80c        2.12" == output[3]


def test_cutting_comments_print_mutate():
    problem = montepy.read_input("tests/inputs/breaking_comments.imcnp")
    cell = problem.cells[1]
    cell.number = 8
    output = cell.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert len(output) == 5
    assert "c this is a cutting comment" == output[3]
    material2 = problem.materials[2]
    material2.number = 5
    output = material2.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert len(output) == 5
    assert "c          26057.80c        2.12" == output[3]


def test_comments_setter(simple_problem):
    cell = copy.deepcopy(simple_problem.cells[1])
    comment = simple_problem.surfaces[1000].comments[0]
    cell.leading_comments = [comment]
    assert cell.comments[0] == comment
    cell.leading_comments = comment
    assert cell.comments[0] == comment
    with pytest.raises(TypeError):
        cell.leading_comments = [5]
    with pytest.raises(TypeError):
        cell.leading_comments = 5


def test_problem_linker():
    cell = montepy.Cell()
    with pytest.raises(TypeError):
        cell.link_to_problem(5)


def test_importance_parsing(importance_problem, simple_problem):
    cell = importance_problem.cells[1]
    assert cell.importance.neutron == 1.0
    assert cell.importance.photon == 1.0
    assert cell.importance.electron == 0.0
    cell = simple_problem.cells[1]
    assert cell.importance.neutron == 1.0
    assert cell.importance.photon == 1.0


def test_importance_format_unmutated(importance_problem):
    imp = importance_problem.cells._importance
    output = imp.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert len(output) == 3
    assert "imp:n,p 1 1 1 0 3" == output[0]
    assert "imp:e   0 2r 1 r" == output[2]


def test_importance_format_mutated(importance_problem):
    problem = copy.deepcopy(importance_problem)
    imp = problem.cells._importance
    problem.cells[1].importance.neutron = 0.5
    with pytest.warns(LineExpansionWarning):
        output = imp.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert len(output) == 4
    assert "imp:n 0.5 1 1 0 3" in output
    assert "c special comment related to #520" == output[2]


def test_importance_write_unmutated(importance_problem):
    fh = io.StringIO()
    importance_problem.write_problem(fh)
    found_np = False
    found_e = False
    fh.seek(0)
    for line in fh.readlines():
        print(line.rstrip())
        if "imp:n,p 1" in line:
            found_np = True
        elif "imp:e" in line:
            found_e = True
    assert found_np
    assert found_e
    fh.close()


def test_importance_write_mutated(importance_problem):
    fh = io.StringIO()
    problem = copy.deepcopy(importance_problem)
    problem.cells[1].importance.neutron = 0.5
    with pytest.warns(LineExpansionWarning):
        problem.write_problem(fh)
    found_n = False
    found_e = False
    fh.seek(0)
    for line in fh.readlines():
        print(line.rstrip())
        if "imp:n 0.5" in line:
            found_n = True
        elif "imp:e" in line:
            found_e = True
    assert found_n
    assert found_e
    fh.close()


def test_importance_write_cell(importance_problem):
    for state in ["no change", "new unmutated cell", "new mutated cell"]:
        fh = io.StringIO()
        problem = copy.deepcopy(importance_problem)
        if "new" in state:
            cell = copy.deepcopy(problem.cells[5])
            cell.number = 999
            problem.cells.append(cell)
        problem.print_in_data_block["imp"] = False
        if "new" in state:
            with pytest.warns(LineExpansionWarning):
                problem.write_problem(fh)
        else:
            problem.write_problem(fh)
        found_np = False
        found_e = False
        found_data_np = False
        fh.seek(0)
        for line in fh.readlines():
            print(line.rstrip())
            if "imp:n,p=1" in line:
                found_np = True
            elif "imp:e=1" in line:
                found_e = True
            elif "imp:e 1" in line.lower():
                found_data_np = True
        assert found_np
        assert found_e
        assert not found_data_np
        fh.close()


def test_importance_write_data(simple_problem):
    fh = io.StringIO()
    problem = copy.deepcopy(simple_problem)
    problem.print_in_data_block["imp"] = True
    problem.write_problem(fh)
    found_n = False
    found_p = False
    fh.seek(0)
    for line in fh:
        print(line.rstrip())
        if "imp:n 1" in line:
            found_n = True
        if "imp:p 1 0.5" in line:
            found_p = True
    assert found_n
    assert found_p
    fh.close()


def test_avoid_blank_cell_modifier_write(simple_problem):
    fh = io.StringIO()
    problem = copy.deepcopy(simple_problem)
    problem.print_in_data_block["U"] = True
    problem.print_in_data_block["FILL"] = True
    problem.print_in_data_block["LAT"] = True
    problem.cells[5].fill.transform = None
    problem.cells[5].fill.universe = None
    problem.cells[1].universe = problem.universes[0]
    for cell in problem.cells:
        del cell.volume
    problem.cells.allow_mcnp_volume_calc = True
    problem.write_problem(fh)
    found_universe = False
    found_lattice = False
    found_vol = False
    found_fill = False
    fh.seek(0)
    for line in fh.readlines():
        print(line.rstrip())
        if "U " in line:
            found_universe = True
        if "LAT " in line:
            found_lattice = True
        if "FILL " in line:
            found_fill = True
        if "VOL " in line:
            found_vol = True
    assert not found_universe
    assert not found_lattice
    assert not found_vol
    assert not found_fill
    fh.close()


def test_set_mode(importance_problem):
    problem = copy.deepcopy(importance_problem)
    problem.set_mode("e p")
    particles = {Particle.ELECTRON, Particle.PHOTON}
    assert len(problem.mode) == 2
    for part in particles:
        assert part in problem.mode


def test_set_equal_importance(importance_problem):
    problem = copy.deepcopy(importance_problem)
    problem.cells.set_equal_importance(0.5, [5])
    for cell in problem.cells:
        for particle in problem.mode:
            if cell.number != 5:
                print(cell.number, particle)
                assert cell.importance[particle] == 0.5
    for particle in problem.mode:
        print(5, particle)
        assert problem.cells[5].importance[particle] == 0.0
    problem.cells.set_equal_importance(0.75, [problem.cells[99]])
    for cell in problem.cells:
        for particle in problem.mode:
            if cell.number != 99:
                print(cell.number, particle)
                assert cell.importance[particle] == 0.75
    for particle in problem.mode:
        print(99, particle)
        assert problem.cells[99].importance[particle] == 0.0
    with pytest.raises(TypeError):
        problem.cells.set_equal_importance("5", [5])
    with pytest.raises(ValueError):
        problem.cells.set_equal_importance(-0.5, [5])
    with pytest.raises(TypeError):
        problem.cells.set_equal_importance(5, "a")
    with pytest.raises(TypeError):
        problem.cells.set_equal_importance(5, ["a"])


def test_check_volume_calculated(simple_problem):
    assert not simple_problem.cells[1].volume_mcnp_calc


def test_redundant_volume():
    with pytest.raises(montepy.errors.MalformedInputError):
        montepy.read_input(os.path.join("tests", "inputs", "test_vol_redundant.imcnp"))


def test_delete_vol(simple_problem):
    problem = copy.deepcopy(simple_problem)
    del problem.cells[1].volume
    assert not problem.cells[1].volume_is_set


def test_enable_mcnp_vol_calc(simple_problem):
    problem = copy.deepcopy(simple_problem)
    problem.cells.allow_mcnp_volume_calc = True
    assert problem.cells.allow_mcnp_volume_calc
    assert "NO" not in str(problem.cells._volume)
    problem.cells.allow_mcnp_volume_calc = False
    assert "NO" in str(problem.cells._volume)
    with pytest.raises(TypeError):
        problem.cells.allow_mcnp_volume_calc = 5


def test_cell_multi_volume():
    in_str = "1 0 -1 VOL=1 VOL 5"
    with pytest.raises(ValueError):
        montepy.Cell(Input([in_str], montepy.input_parser.block_type.BlockType.CELL))


def test_universe_cell_parsing(simple_problem):
    answers = [350] + [0] * 4
    for cell, answer in zip(simple_problem.cells, answers):
        print(cell, answer)
        assert cell.universe.number == answer


def test_universe_fill_data_parsing(data_universe_problem):
    answers = [350, 0, 0, 1]
    for cell, answer in zip(data_universe_problem.cells, answers):
        print(cell, answer)
        assert cell.universe.number == answer
    for cell in data_universe_problem.cells:
        print(cell)
        if cell.number != 99:
            assert not cell.not_truncated
        else:
            assert cell.not_truncated
    assert data_universe_problem.cells[99].not_truncated
    answers = [None, None, 350, None, None]
    for cell, answer in zip(data_universe_problem.cells, answers):
        print(cell.number, cell.fill.universe, answer)
        if answer is None:
            assert cell.fill.universe is None
        else:
            assert cell.fill.universe.number == answer


def test_universe_cells1(data_universe_problem):
    answers = {350: [1], 0: [2, 3, 5], 1: [99]}
    for uni_number, cell_answers in answers.items():
        for cell, answer in zip(
            data_universe_problem.universes[uni_number].cells, cell_answers
        ):
            assert cell.number == answer


def test_cell_not_truncate_setter(simple_problem):
    problem = copy.deepcopy(simple_problem)
    cell = problem.cells[1]
    cell.not_truncated = True
    assert cell.not_truncated
    with pytest.raises(ValueError):
        cell = problem.cells[2]
        cell.not_truncated = True


def test_universe_setter(simple_problem):
    problem = copy.deepcopy(simple_problem)
    universe = problem.universes[350]
    cell = problem.cells[3]
    cell.universe = universe
    assert cell.universe == universe
    assert cell.universe.number == 350
    with pytest.raises(TypeError):
        cell.universe = 5


def test_universe_cell_formatter(simple_problem):
    problem = copy.deepcopy(simple_problem)
    universe = problem.universes[350]
    cell = problem.cells[3]
    cell.universe = universe
    cell.not_truncated = True
    with pytest.warns(LineExpansionWarning):
        output = cell.format_for_mcnp_input((6, 2, 0))
    assert "U=-350" in " ".join(output)


def test_universe_data_formatter(data_universe_problem):
    problem = copy.deepcopy(data_universe_problem)
    # test unmutated
    output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert "u 350 2J -1" in output
    universe = problem.universes[350]
    # test mutated
    cell = problem.cells[3]
    cell.universe = universe
    cell.not_truncated = True
    with pytest.warns(LineExpansionWarning):
        output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert "u 350 J -350 -1" in output
    # test appending a new mutated cell
    new_cell = copy.deepcopy(cell)
    new_cell.number = 1000
    new_cell.universe = universe
    new_cell.not_truncated = False
    problem.cells.append(new_cell)
    with pytest.warns(LineExpansionWarning):
        output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert "u 350 J -350 -1 J 350 " in output
    # test appending a new UNmutated cell
    problem = copy.deepcopy(data_universe_problem)
    cell = problem.cells[3]
    new_cell = copy.deepcopy(cell)
    new_cell.number = 1000
    new_cell.universe = universe
    new_cell.not_truncated = False
    # lazily implement pulling cell in from other model
    new_cell._mutated = False
    new_cell._universe._mutated = False
    problem.cells.append(new_cell)
    with pytest.warns(LineExpansionWarning):
        output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
    print(output)
    assert "u 350 2J -1 J 350 " in output


def test_universe_number_collision():
    problem = montepy.read_input(
        os.path.join("tests", "inputs", "test_universe_data.imcnp")
    )
    with pytest.raises(montepy.errors.NumberConflictError):
        problem.universes[0].number = 350

    with pytest.raises(ValueError):
        problem.universes[350].number = 0


def test_universe_repr(simple_problem):
    uni = simple_problem.universes[0]
    output = repr(uni)
    assert "Number: 0" in output
    assert "Problem: set" in output
    assert "Cells: [2" in output


def test_lattice_format_data(simple_problem):
    problem = copy.deepcopy(simple_problem)
    cells = problem.cells
    cells[1].lattice = 1
    cells[99].lattice = 2
    answer = "LAT 1 2J 2"
    output = cells._lattice.format_for_mcnp_input((6, 2, 0))
    assert answer in output[0]


def test_lattice_push_to_cells(simple_problem):
    problem = copy.deepcopy(simple_problem)
    lattices = [1, 2, Jump(), Jump()]
    card = Input(
        ["Lat " + " ".join(list(map(str, lattices)))],
        montepy.input_parser.block_type.BlockType.DATA,
    )
    lattice = montepy.data_inputs.lattice_input.LatticeInput(card)
    lattice.link_to_problem(problem)
    lattice.push_to_cells()
    for cell, answer in zip(problem.cells, lattices):
        print(cell.number, answer)
        if isinstance(answer, int):
            assert cell.lattice.value == answer
        else:
            assert cell.lattice is None


def test_universe_problem_parsing(universe_problem):
    for cell in universe_problem.cells:
        if cell.number == 1:
            assert cell.universe.number == 1
        else:
            assert cell.universe.number == 0


def test_importance_end_repeat(universe_problem):
    problem = copy.deepcopy(universe_problem)
    for cell in problem.cells:
        if cell.number in {99, 5}:
            cell.importance.photon = 1.0
        else:
            cell.importance.photon = 0.0
    problem.print_in_data_block["IMP"] = True
    output = problem.cells._importance.format_for_mcnp_input((6, 2, 0))
    # OG value was 0.5 so 0.0 is correct.
    assert "imp:p 0 0.0" in output


def test_fill_parsing(universe_problem):
    answers = [None, np.array([[[1], [0]], [[0], [1]]]), None, 1, 1]
    for cell, answer in zip(universe_problem.cells, answers):
        if answer is None:
            assert cell.fill.universe is None
        elif isinstance(answer, np.ndarray):
            assert cell.fill.multiple_universes
            assert (cell.fill.min_index == np.array([0.0, 0.0, 0.0])).all()
            assert (cell.fill.max_index == np.array([1.0, 1.0, 0.0])).all()
            assert cell.fill.universes[0][0][0].number == answer[0][0][0]
            assert cell.fill.universes[1][1][0].number == answer[1][1][0]
            assert cell.fill.transform == universe_problem.transforms[5]
        else:
            assert cell.fill.universe.number == answer


def test_fill_transform_setter(universe_problem):
    problem = copy.deepcopy(universe_problem)
    transform = problem.transforms[5]
    cell = problem.cells[5]
    cell.fill.transform = transform
    assert cell.fill.transform == transform
    assert not cell.fill.hidden_transform
    cell.fill.transform = None
    assert cell.fill.transform is None
    with pytest.raises(TypeError):
        cell.fill.transform = "hi"
    cell.fill.transform = transform
    del cell.fill.transform
    assert cell.fill.transform is None


def test_fill_cell_format(simple_problem, universe_problem):
    problem = copy.deepcopy(universe_problem)
    fill = problem.cells[5].fill
    output = fill.format_for_mcnp_input((6, 2, 0))
    answer = "fill=1 (1 0.0 0.0)"
    assert output[0] == answer
    # test *fill
    fill.transform.is_in_degrees = True
    output = fill.format_for_mcnp_input((6, 2, 0))
    answer = "*fill=1 (1 0.0 0.0)"
    assert output[0] == answer
    # test changing the transform
    fill.transform.displacement_vector[0] = 2.0
    output = fill.format_for_mcnp_input((6, 2, 0))
    assert output[0] == "*fill=1 (2 0.0 0.0)"
    # test without transform
    fill.transform = None
    answer = "fill=1 "
    output = fill.format_for_mcnp_input((6, 2, 0))
    assert output[0] == answer
    # test with no fill
    fill.universe = None
    output = fill.format_for_mcnp_input((6, 2, 0))
    assert len(output) == 0
    # test with complex universe lattice fill
    fill = problem.cells[2].fill
    output = fill.format_for_mcnp_input((6, 2, 0))
    answers = ["fill= 0:1 0:1 0:0 1 0 0 1 (5)"]
    assert output == answers
    problem.print_in_data_block["FILL"] = True
    # test that complex fill is not printed in data block
    with pytest.raises(ValueError):
        problem.cells._fill.format_for_mcnp_input((6, 2, 0))
    problem = copy.deepcopy(simple_problem)
    problem.cells[5].fill.transform = None
    problem.print_in_data_block["FILL"] = True
    output = problem.cells._fill.format_for_mcnp_input((6, 2, 0))
    assert output == ["FILL 4J 350 "]


def test_universe_cells_claim(universe_problem):
    problem = copy.deepcopy(universe_problem)
    universe = problem.universes[1]
    universe.claim(problem.cells[2])
    assert problem.cells[2].universe == universe
    universe = montepy.Universe(5)
    problem.universes.append(universe)
    universe.claim(problem.cells)
    for cell in problem.cells:
        assert cell.universe == universe
    with pytest.raises(TypeError):
        universe.claim("hi")
    with pytest.raises(TypeError):
        universe.claim(["hi"])


def test_universe_cells2(universe_problem):
    answers = [1]
    universe = universe_problem.universes[1]
    assert len(answers) == len(list(universe.cells))
    for cell, answer in zip(universe.cells, answers):
        assert cell.number == answer


def test_data_print_control_str(simple_problem):
    assert (
        str(simple_problem.print_in_data_block)
        == "Print data in data block: {'imp': False, 'u': False, 'fill': False, 'vol': True}"
    )


def test_cell_validator(simple_problem):
    problem = copy.deepcopy(simple_problem)
    cell = problem.cells[1]
    del cell.mass_density
    with pytest.raises(montepy.errors.IllegalState):
        cell.validate()
    cell = montepy.Cell()
    # test no geometry at all
    with pytest.raises(montepy.errors.IllegalState):
        cell.validate()
    surf = problem.surfaces[1000]
    cell.surfaces.append(surf)
    # test surface added but geomtry not defined
    with pytest.raises(montepy.errors.IllegalState):
        cell.validate()


def test_importance_rewrite(simple_problem):
    out_file = "test_import_data_1"
    problem = copy.deepcopy(simple_problem)
    problem.print_in_data_block["imp"] = True
    try:
        problem.write_problem(out_file)
        problem = montepy.read_input(out_file)
        os.remove(out_file)
        problem.print_in_data_block["imp"] = False
        problem.write_problem(out_file)
        found_n = False
        found_p = False
        found_vol = False
        with open(out_file, "r") as fh:
            for line in fh:
                print(line.rstrip())
                if "IMP:N 1 2R" in line:
                    found_n = True
                if "IMP:P 1 0.5" in line:
                    found_p = True
                if "vol NO 2J 1 1.5 J" in line:
                    found_vol = True
        assert not found_n
        assert not found_p
        assert found_vol
    finally:
        try:
            os.remove(out_file)
        except FileNotFoundError:
            pass


def test_parsing_error():
    in_file = os.path.join("tests", "inputs", "test_bad_syntax.imcnp")
    with pytest.raises(montepy.errors.ParsingError):
        problem = montepy.read_input(in_file)


def test_leading_comments(simple_problem):
    cell = copy.deepcopy(simple_problem.cells[1])
    leading_comments = cell.leading_comments
    assert "cells" in leading_comments[0].contents
    del cell.leading_comments
    assert not cell.leading_comments
    cell.leading_comments = leading_comments[0:1]
    assert "cells" in cell.leading_comments[0].contents
    assert len(cell.leading_comments) == 1


def test_wrap_warning(simple_problem):
    cell = copy.deepcopy(simple_problem.cells[1])
    with pytest.warns(montepy.errors.LineExpansionWarning):
        output = cell.wrap_string_for_mcnp("h" * 130, (6, 2, 0), True)
        assert len(output) == 2
    output = cell.wrap_string_for_mcnp("h" * 127, (6, 2, 0), True)
    assert len(output) == 1


def test_expansion_warning_crash(simple_problem):
    problem = copy.deepcopy(simple_problem)
    cell = problem.cells[99]
    cell.material = problem.materials[1]
    cell.mass_density = 10.0
    problem.materials[1].number = 987654321
    problem.surfaces[1010].number = 123456789
    with io.StringIO() as fh:
        with pytest.warns(montepy.errors.LineExpansionWarning):
            problem.write_problem(fh)


def test_alternate_encoding():
    with pytest.raises(UnicodeDecodeError):
        montepy.read_input(
            os.path.join("tests", "inputs", "bad_encoding.imcnp"), replace=False
        )
    montepy.read_input(
        os.path.join("tests", "inputs", "bad_encoding.imcnp"), replace=True
    )


_SKIP_LINES = {
    # skip lines of added implied importances
    "tests/inputs/test_universe_data.imcnp": {5: True, 14: True, 15: True},
}


@pytest.mark.parametrize(
    "file",
    set((Path("tests") / "inputs").iterdir())
    - {
        Path("tests")
        / "inputs"
        / p  #                           Skip complexity of read
        for p in constants.BAD_INPUTS
        | constants.IGNORE_FILES
        | {"testRead.imcnp", "readEdgeCase.imcnp"}
    },
)
def test_read_write_cycle(file):
    print(f"Testing against {file} *********************")
    problem = montepy.read_input(file)
    SKIPPERS = _SKIP_LINES.get(str(file), {})
    fh = io.StringIO()
    problem.write_problem(fh)
    fh.seek(0)
    # test valid syntax
    new_problem = montepy.MCNP_Problem("foo")
    # verify lines are similar
    fh.seek(0)
    lines = [line.rstrip() for line in fh]
    [print(line) for line in lines]
    with open(file, "r") as gold_fh:
        gold_fh_iter = iter(gold_fh)
        lines_iter = iter(lines)
        for i, (gold_line, new_line) in enumerate(zip(gold_fh_iter, lines_iter)):
            if i in SKIPPERS:
                # True means skip new file line
                if SKIPPERS[i]:
                    new_line = next(lines_iter)
                else:
                    gold_line = next(gold_fh_iter)
            # edge case override for not fixing #527.
            if str(file) == "tests/inputs/test_interp_edge.imcnp" and i == 1:
                assert new_line == "10214   0    (1  2I 4 )"
                continue
            try:
                assert new_line == gold_line.rstrip().expandtabs(8)
            except AssertionError as e:
                # handle case of making importance explicit
                if "IMP:n=0.0" in new_line:
                    assert (
                        new_line.replace("IMP:n=0.0", "").rstrip() == gold_line.rstrip()
                    )
                else:
                    raise e



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.data_inputs.cell_modifier import CellModifierInput
from montepy.errors import *
from montepy.constants import DEFAULT_VERSION
from montepy.input_parser.mcnp_input import Jump
from montepy.input_parser import syntax_node
from montepy.mcnp_object import MCNP_Object
from montepy.utilities import *


def _ensure_positive(self, value):
    if value < 0:
        raise ValueError(f"Volume must be positive. {value} given.")


class Volume(CellModifierInput):
    """
    Class for the data input that modifies cell volumes; ``VOL``.

    :param input: the Input object representing this data input
    :type input: Input
    :param in_cell_block: if this card came from the cell block of an input file.
    :type in_cell_block: bool
    :param key: the key from the key-value pair in a cell
    :type key: str
    :param value: the value syntax tree from the key-value pair in a cell
    :type value: SyntaxNode
    """

    def __init__(self, input=None, in_cell_block=False, key=None, value=None):
        self._volume = self._generate_default_node(float, None)
        self._calc_by_mcnp = True
        super().__init__(input, in_cell_block, key, value)
        if self.in_cell_block:
            if key:
                value = self._tree["data"][0]
                if value.type != float or value.value < 0:
                    raise ValueError(
                        f"Cell volume must be a number ≥ 0.0. {value} was given"
                    )
                self._volume = value
                self._calc_by_mcnp = False
        elif input:
            self._volume = []
            tree = self._tree
            if "parameters" in tree:
                raise MalformedInputError(
                    input, f"Volume card can't accept any key-value parameters"
                )
            if (
                "keyword" in tree
                and tree["keyword"].value
                and tree["keyword"].value.lower() == "no"
            ):
                self._calc_by_mcnp = False
            for node in tree["data"]:
                if node.value is not None:
                    try:
                        assert node.type == float
                        assert node.value >= 0
                        self._volume.append(node)
                    except AssertionError:
                        raise MalformedInputError(
                            input, f"Cell volumes by a number ≥ 0.0: {node} given"
                        )
                else:
                    self._volume.append(node)

    def _generate_default_cell_tree(self):
        list_node = syntax_node.ListNode("number sequence")
        list_node.append(self._generate_default_node(float, None))
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = self._generate_default_node(
            str, self._class_prefix().upper(), None
        )
        self._tree = syntax_node.SyntaxNode(
            "volume",
            {
                "classifier": classifier,
                "param_seperator": self._generate_default_node(str, "=", None),
                "data": list_node,
            },
        )

    @staticmethod
    def _class_prefix():
        return "vol"

    @staticmethod
    def _has_number():
        return False

    @staticmethod
    def _has_classifier():
        return 0

    @make_prop_val_node(
        "_volume",
        (float, int, type(None)),
        float,
        validator=_ensure_positive,
        deletable=True,
    )
    def volume(self):
        """
        The actual cell volume.

        Only available at the cell level.

        :returns: the cell volume iff this is for a single cell
        :rtype: float
        """
        pass

    @property
    def _tree_value(self):
        if self.in_cell_block:
            return self._volume

    @property
    def is_mcnp_calculated(self):
        """
        Indicates whether or not the cell volume will attempt to be calculated by MCNP.

        This can be disabled by either manually setting the volume or disabling
        this calculation globally.
        This does not guarantee that MCNP will able to do so.
        Complex geometries may make this impossible.

        :returns: True iff MCNP will try to calculate the volume for this cell.
        :rtype: bool
        """
        if self._problem and self.in_cell_block:
            if not self._problem.cells._volume.is_mcnp_calculated:
                return False
        return self._calc_by_mcnp

    @is_mcnp_calculated.setter
    def is_mcnp_calculated(self, value):
        if not self.in_cell_block:
            self._calc_by_mcnp = value

    @property
    def has_information(self):
        if self.in_cell_block:
            return self.set

    @property
    def set(self) -> bool:
        """
        If this volume is set.

        :returns: true if the volume is manually set.
        :rtype: bool
        """
        return self.volume is not None

    def merge(self, other):
        raise MalformedInputError(
            other._input, "Cannot have two volume inputs for the problem"
        )

    def push_to_cells(self):
        if not self.in_cell_block and self._problem and self._volume:
            self._check_redundant_definitions()
            cells = self._problem.cells
            for i, cell in enumerate(cells):
                if i >= len(self._volume):
                    return
                vol = self._volume[i]
                if not isinstance(vol, Jump):
                    cell._volume._volume = vol

    def _clear_data(self):
        del self._volume

    def __str__(self):
        ret = "\n".join(self.format_for_mcnp_input(DEFAULT_VERSION))
        return ret

    def __repr__(self):
        ret = (
            f"VOLUME: in_cell: {self._in_cell_block}, calc_by_mcnp: {self.is_mcnp_calculated},"
            f" set_in_block: {self.set_in_cell_block}, "
            f"Volume : {self._volume if hasattr(self, '_volume') else ''}"
        )
        return ret

    def _update_values(self):
        if self.in_cell_block:
            self._update_cell_values()
        else:
            keyword = self._tree["keyword"]
            if not self.is_mcnp_calculated and (
                keyword.value is None or keyword.value.lower() != "no"
            ):
                keyword.value = "NO"
                if keyword.padding is None or len(keyword.padding) == 0:
                    keyword.padding = syntax_node.PaddingNode(" ")
            elif self.is_mcnp_calculated:
                keyword.value = None
            new_vals = self._collect_new_values()
            self.data.update_with_new_values(new_vals)

    def _update_cell_values(self):
        if self._tree["data"][0] is not self._volume:
            self._tree["data"].nodes.pop()
            self._tree["data"].append(self._volume)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum, unique


@unique
class Lattice(Enum):
    """
    Represents the options for the lattice ``LAT``.
    """

    HEXAHEDRA = 1
    """
    Hexhedra are solids with six faces.

    One such solid is a rectangular prism.
    """
    HEXAGONAL = 2
    """
    Hexagonal prism are solids with eight faces.
    """



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import abstractmethod
from montepy.mcnp_object import MCNP_Object


class Numbered_MCNP_Object(MCNP_Object):
    @property
    @abstractmethod
    def number(self):
        """
        The current number of the object that will be written out to a new input.

        :rtype: int
        """
        pass

    @property
    @abstractmethod
    def old_number(self):
        """
        The original number of the object provided in the input file

        :rtype: int
        """
        pass



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from montepy import mcnp_object
from montepy.data_inputs import data_input
from montepy.errors import *
from montepy.numbered_mcnp_object import Numbered_MCNP_Object
from montepy.utilities import *
import numpy as np
import re


def _enforce_number(self, val):
    if val <= 0:
        raise ValueError(f"Transform number must be > 0. {val} given.")


class Transform(data_input.DataInputAbstract, Numbered_MCNP_Object):
    """
    Input to represent a transform input (TR).

    :param input: The Input syntax object this will wrap and parse.
    :type input: Input
    """

    def __init__(self, input=None, pass_through=False):
        self._pass_through = pass_through
        self._number = self._generate_default_node(int, -1)
        self._old_number = self._generate_default_node(int, -1)
        self._displacement_vector = np.array([])
        self._rotation_matrix = np.array([])
        self._is_in_degrees = False
        self._is_main_to_aux = True
        super().__init__(input)
        if input:
            words = self._tree["data"]
            i = 0
            if len(words) < 3:
                raise MalformedInputError(input, f"Not enough entries were provided")
            modifier = self._classifier.modifier
            if modifier and "*" in modifier.value:
                self._is_in_degrees = True
            else:
                self._is_in_degrees = False
            self._number = self._input_number
            self._old_number = copy.deepcopy(self._number)

            # parse displacement
            values = []
            for j, word in enumerate(words):
                values.append(word.value)
                i += 1
                if j >= 2:
                    break
            self._displacement_vector = np.array(values)

            # parse rotation
            values = []
            for j, word in enumerate(words.nodes[i:]):
                values.append(word.value)
                i += 1
                if j >= 8:
                    break
            self._rotation_matrix = np.array(values)

            self._is_main_to_aux = True
            if len(values) == 9:
                try:
                    word = words[i]
                    word.is_negatable_identifier = True
                    if word.value != 1:
                        raise MalformedInputError(
                            input, f"{word} can't be parsed as 1 or -1"
                        )
                    # negative means not main_to_aux
                    self._is_main_to_aux = not word.is_negative
                # if no more words remain don't worry
                except IndexError:
                    pass

    @staticmethod
    def _class_prefix():
        return "tr"

    @staticmethod
    def _has_number():
        return True

    @staticmethod
    def _has_classifier():
        return 0

    @property
    def hidden_transform(self):
        """
        Whether or not this transform is "hidden" i.e., has no number.

        If True this transform was created from a fill card, and has no number.

        :rtype: bool
        """
        return self._pass_through

    @make_prop_pointer("_is_in_degrees", bool)
    def is_in_degrees(self):
        """
        The rotation matrix is in degrees and not in cosines

        :rtype: bool
        """
        pass

    @make_prop_val_node("_number", (int, float), int, _enforce_number)
    def number(self):
        """
        The transform number for this transform

        :rtype: int

        """
        pass

    @make_prop_val_node("_old_number")
    def old_number(self):
        """
        The transform number used in the original file

        :rtype: int
        """
        pass

    @property
    def displacement_vector(self):
        """
        The transform displacement vector

        :rtype: numpy.array
        """
        return self._displacement_vector

    @displacement_vector.setter
    def displacement_vector(self, vector):
        if not isinstance(vector, np.ndarray):
            raise TypeError("displacement_vector must be a numpy array")
        if len(vector) != 3:
            raise ValueError("displacement_vector must have three components")
        self._displacement_vector = vector

    @property
    def rotation_matrix(self):
        """
        The rotation matrix

        :rtype: np.array
        """
        return self._rotation_matrix

    @rotation_matrix.setter
    def rotation_matrix(self, matrix):
        if not isinstance(matrix, np.ndarray):
            raise TypeError("rotation_matrix must be a numpy array")
        if len(matrix) < 5 or len(matrix) > 9:
            raise ValueError("rotation_matrix must have between 5 and 9 components.")
        self._rotation_matrix = matrix

    @make_prop_pointer("_is_main_to_aux", bool)
    def is_main_to_aux(self):
        """
        Whether or not the displacement vector points from the main origin to auxilary
        origin, or vice versa.

        :rtype: bool
        """
        return self._is_main_to_aux

    def __str__(self):
        return f"TRANSFORM: {self.number}"

    def __repr__(self):
        ret = f"TRANSFORM: {self.number}\n"
        ret += f"DISPLACE: {self.displacement_vector}\n"
        ret += f"ROTATE: {self.rotation_matrix}\n"
        ret += f"MAIN_TO_AUX: {self.is_main_to_aux}\n"
        return ret

    def _update_values(self):
        # update in degrees
        if self._classifier.modifier is None:
            self._classifier.modifier = self._generate_default_node(
                str, "", padding=None
            )
        if self.is_in_degrees:
            self._classifier.modifier.value = "*"
        else:
            self._classifier.modifier.value = ""
        # update displacement vector
        new_values = []
        list_iter = iter(self.data)
        length = len(self.data)
        for value, node in zip(self.displacement_vector, list_iter):
            node.value = value
            new_values.append(node)
        # update the rotation matrix
        # test if the rotation matrix has info, or was specified or main_to_aux is needed
        needs_rotation = (
            np.any(self.rotation_matrix)
            or len(self.data) >= 8
            or not self.is_main_to_aux
        )
        if needs_rotation:
            flat_pack = self.rotation_matrix
            i = -1
            for i, (value, node) in enumerate(zip(flat_pack, list_iter)):
                node.value = value
                new_values.append(node)
            if i < len(flat_pack) - 1:
                for value in flat_pack[i + 1 :]:
                    node = self._generate_default_node(float, value)
                    self.data.append(node)
                    new_values.append(node)
            # if main to aux specified or is needed
            if len(self.data) == 13 or not self.is_main_to_aux:
                if len(self.data) == 13:
                    node = self.data[-1]
                else:
                    node = self._generate_default_node(int, 1)
                    node.is_negatable_identifier = True
                    self.data.append(node)
                node.is_negative = not self.is_main_to_aux
                new_values.append(node)
        # Trigger shortcut recompression
        self.data.update_with_new_values(new_values)

    def validate(self):
        if self.displacement_vector is None or len(self.displacement_vector) != 3:
            raise IllegalState(
                f"Transform: {self.number} does not have a valid displacement Vector"
            )

    def equivalent(self, other, tolerance):
        """Determines if this is effectively equivalent to another transformation

        :param other: The transform to compare self again.
        :type other: Transform
        :param tolerance: the allowable difference in any attribute to still be considered equivalent.
        :type tolerance: float

        :returns: True iff all transform elements in both are within the tolerance of each other.
        :rtype: bool
        """

        if self.is_in_degrees != other.is_in_degrees:
            return False

        if self.is_main_to_aux != other.is_main_to_aux:
            return False

        for i, component in enumerate(self.displacement_vector):
            if abs(component - other.displacement_vector[i]) >= tolerance:
                return False

        if len(self.rotation_matrix) > 0:
            if len(other.rotation_matrix) == 0:
                return False
            for i, component in enumerate(self.rotation_matrix):
                if abs(component - other.rotation_matrix[i]) >= tolerance:
                    return False
        return True



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import ABC, abstractmethod
import itertools as it
from montepy.errors import *
from montepy.constants import (
    BLANK_SPACE_CONTINUE,
    get_max_line_length,
    rel_tol,
    abs_tol,
)
from montepy.input_parser.syntax_node import (
    CommentNode,
    PaddingNode,
    ParametersNode,
    ValueNode,
)
import montepy
import numpy as np
import textwrap
import warnings


class MCNP_Object(ABC):
    """
    Abstract class for semantic representations of MCNP inputs.

    .. versionchanged:: 0.2.0
        Generally significant changes for parser rework.
        For init removed ``comments``, and added ``parser`` as arguments.

    :param input: The Input syntax object this will wrap and parse.
    :type input: Input
    :param parser: The parser object to parse the input with.
    :type parser: MCNP_Lexer
    """

    def __init__(self, input, parser):
        self._problem = None
        self._parameters = ParametersNode()
        self._input = None
        if input:
            if not isinstance(input, montepy.input_parser.mcnp_input.Input):
                raise TypeError("input must be an Input")
            try:
                try:
                    parser.restart()
                # raised if restarted without ever parsing
                except AttributeError as e:
                    pass
                self._tree = parser.parse(input.tokenize(), input)
                self._input = input
            except ValueError as e:
                raise MalformedInputError(
                    input, f"Error parsing object of type: {type(self)}: {e.args[0]}"
                )
            if self._tree is None:
                raise ParsingError(
                    input,
                    "",
                    parser.log.clear_queue(),
                )
            if "parameters" in self._tree:
                self._parameters = self._tree["parameters"]

    @staticmethod
    def _generate_default_node(value_type, default, padding=" "):
        """
        Generates a "default" or blank ValueNode.

        None is generally a safe default value to provide.

        .. versionadded:: 0.2.0

        :param value_type: the data type for the ValueNode.
        :type value_type: Class
        :param default: the default value to provide (type needs to agree with value_type)
        :type default: value_type
        :param padding: the string to provide to the PaddingNode. If None no PaddingNode will be added.
        :type padding: str, None
        :returns: a new ValueNode with the requested information.
        :rtype: ValueNode
        """
        if padding:
            padding_node = PaddingNode(padding)
        else:
            padding_node = None
        if default is None or isinstance(default, montepy.input_parser.mcnp_input.Jump):
            return ValueNode(default, value_type, padding_node)
        return ValueNode(str(default), value_type, padding_node)

    @property
    def parameters(self):
        """
        A dictionary of the additional parameters for the object.

        e.g.: ``1 0 -1 u=1 imp:n=0.5`` has the parameters
        ``{"U": "1", "IMP:N": "0.5"}``

        :returns: a dictionary of the key-value pairs of the parameters.
        :rytpe: dict
        """
        return self._parameters

    @abstractmethod
    def _update_values(self):
        """
        Method to update values in syntax tree with new values.

        Generally when :func:`~montepy.utilities.make_prop_val_node` this is not necessary to do,
        but when :func:`~montepy.utilities.make_prop_pointer` is used it is necessary.
        The most common need is to update a value based on the number for an object pointed at,
        e.g., the material number in a cell definition.

        .. versionadded:: 0.2.0

        """
        pass

    def format_for_mcnp_input(self, mcnp_version):
        """
        Creates a string representation of this MCNP_Object that can be
        written to file.

        :param mcnp_version: The tuple for the MCNP version that must be exported to.
        :type mcnp_version: tuple
        :return: a list of strings for the lines that this input will occupy.
        :rtype: list
        """
        self.validate()
        self._update_values()
        self._tree.check_for_graveyard_comments()
        lines = self.wrap_string_for_mcnp(self._tree.format(), mcnp_version, True)
        return lines

    @property
    def comments(self):
        """
        The comments associated with this input if any.

        This includes all ``C`` comments before this card that aren't part of another card,
        and any comments that are inside this card.

        :returns: a list of the comments associated with this comment.
        :rtype: list
        """
        return list(self._tree.comments)

    @property
    def leading_comments(self):
        """
        Any comments that come before the beginning of the input proper.

        .. versionadded:: 0.2.0

        :returns: the leading comments.
        :rtype: list
        """
        return list(self._tree["start_pad"].comments)

    @leading_comments.setter
    def leading_comments(self, comments):
        if not isinstance(comments, (list, tuple, CommentNode)):
            raise TypeError(
                f"Comments must be a CommentNode, or a list of Comments. {comments} given."
            )
        if isinstance(comments, CommentNode):
            comments = [comments]
        for i, comment in enumerate(comments):
            if not isinstance(comment, CommentNode):
                raise TypeError(
                    f"Comment must be a CommentNode. {comment} given at index {i}."
                )
        new_nodes = list(*zip(comments, it.cycle(["\n"])))
        if self._tree["start_pad"] is None:
            self._tree["start_pad"] = syntax_node.PaddingNode(" ")
        self._tree["start_pad"]._nodes = new_nodes

    @leading_comments.deleter
    def leading_comments(self):
        self._tree["start_pad"]._delete_trailing_comment()

    @staticmethod
    def wrap_string_for_mcnp(string, mcnp_version, is_first_line):
        """
        Wraps the list of the words to be a well formed MCNP input.

        multi-line inputs will be handled by using the indentation format,
        and not the "&" method.

        :param string: A long string with new lines in it,
                    that needs to be chunked appropriately for MCNP inputs
        :type string: str
        :param mcnp_version: the tuple for the MCNP that must be formatted for.
        :type mcnp_version: tuple
        :param is_first_line: If true this will be the beginning of an MCNP input.
                             The first line will not be indented.
        :type is_first_line: bool
        :returns: A list of strings that can be written to an input file, one item to a line.
        :rtype: list
        """
        line_length = get_max_line_length(mcnp_version)
        indent_length = BLANK_SPACE_CONTINUE
        strings = string.splitlines()
        if is_first_line:
            initial_indent = 0
        else:
            initial_indent = indent_length
        wrapper = textwrap.TextWrapper(
            width=line_length,
            initial_indent=" " * initial_indent,
            subsequent_indent=" " * indent_length,
            drop_whitespace=False,
        )
        ret = []
        for line in strings:
            buffer = wrapper.wrap(line)
            if len(buffer) > 1:
                warning = LineExpansionWarning(
                    f"The line exceeded the maximum length allowed by MCNP, and was split. The line was:\n{line}"
                )
                warning.cause = "line"
                warning.og_value = line
                warning.new_value = buffer
                warnings.warn(
                    warning,
                    LineExpansionWarning,
                    stacklevel=2,
                )
            ret += buffer
        return ret

    def validate(self):
        """
        Validates that the object is in a usable state.

        :raises: IllegalState if any condition exists that make the object incomplete.
        """
        pass

    def link_to_problem(self, problem):
        """Links the input to the parent problem for this input.

        This is done so that inputs can find links to other objects.

        :param problem: The problem to link this input to.
        :type problem: MCNP_Problem
        """
        if not isinstance(problem, montepy.mcnp_problem.MCNP_Problem):
            raise TypeError("problem must be an MCNP_Problem")
        self._problem = problem

    @property
    def trailing_comment(self):
        """
        The trailing comments and padding of an input.

        Generally this will be blank as these will be moved to be a leading comment for the next input.

        :returns: the trailing ``c`` style comments and intermixed padding (e.g., new lines)
        :rtype: list
        """
        return self._tree.get_trailing_comment()

    def _delete_trailing_comment(self):
        self._tree._delete_trailing_comment()

    def _grab_beginning_comment(self, padding):
        if padding:
            self._tree["start_pad"]._grab_beginning_comment(padding)

    @staticmethod
    def wrap_words_for_mcnp(words, mcnp_version, is_first_line):  # pragma: no cover
        """
        Wraps the list of the words to be a well formed MCNP input.

        multi-line cards will be handled by using the indentation format,
        and not the "&" method.

        .. deprecated:: 0.2.0
            The concept of words is deprecated, and should be handled by syntax trees now.

        :param words: A list of the "words" or data-grams that needed to added to this card.
                      Each word will be separated by at least one space.
        :type words: list
        :param mcnp_version: the tuple for the MCNP that must be formatted for.
        :type mcnp_version: tuple
        :param is_first_line: If true this will be the beginning of an MCNP card.
                             The first line will not be indented.
        :type is_first_line: bool
        :returns: A list of strings that can be written to an input file, one item to a line.
        :rtype: list
        :raises DeprecationWarning: raised always.
        """
        warnings.warn(
            "wrap_words_for_mcnp is deprecated. Use syntax trees instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        string = " ".join(words)
        return MCNP_Card.wrap_string_for_mcnp(string, mcnp_version, is_first_line)

    @staticmethod
    def compress_repeat_values(values, threshold=1e-6):  # pragma: no cover
        """
        Takes a list of floats, and tries to compress it using repeats.

        E.g., 1 1 1 1 would compress to 1 3R

        .. deprecated:: 0.2.0
            This should be automatically handled by the syntax tree instead.

        :param values: a list of float values to try to compress
        :type values: list
        :param threshold: the minimum threshold to consider two values different
        :type threshold: float
        :returns: a list of MCNP word strings that have repeat compression
        :rtype: list
        :raises DeprecationWarning: always raised.
        """
        warnings.warn(
            "compress_repeat_values is deprecated, and shouldn't be necessary anymore",
            DeprecationWarning,
            stacklevel=2,
        )
        ret = []
        last_value = None
        float_formatter = "{:n}"
        repeat_counter = 0

        def flush_repeats():
            nonlocal repeat_counter, ret
            if repeat_counter >= 2:
                ret.append(f"{repeat_counter}R")
            elif repeat_counter == 1:
                ret.append(float_formatter.format(last_value))
            repeat_counter = 0

        for value in values:
            if isinstance(value, montepy.input_parser.mcnp_input.Jump):
                ret.append(value)
                last_value = None
            elif last_value:
                if np.isclose(value, last_value, atol=threshold):
                    repeat_counter += 1
                else:
                    flush_repeats()
                    ret.append(float_formatter.format(value))
                    last_value = value
            else:
                ret.append(float_formatter.format(value))
                last_value = value
                repeat_counter = 0
        flush_repeats()
        return ret

    @staticmethod
    def compress_jump_values(values):  # pragma: no cover
        """
        Takes a list of strings and jump values and combines repeated jump values.

        e.g., 1 1 J J 3 J becomes 1 1 2J 3 J

        .. deprecated:: 0.2.0
            This should be automatically handled by the syntax tree instead.

        :param values: a list of string and Jump values to try to compress
        :type values: list
        :returns: a list of MCNP word strings that have jump compression
        :rtype: list
        :raises DeprecationWarning: raised always.
        """
        warnings.warn(
            "compress_jump_values is deprecated, and will be removed in the future.",
            DeprecationWarning,
            stacklevel=2,
        )
        ret = []
        jump_counter = 0

        def flush_jumps():
            nonlocal jump_counter, ret
            if jump_counter == 1:
                ret.append("J")
            elif jump_counter >= 1:
                ret.append(f"{jump_counter}J")
            jump_counter = 0

        for value in values:
            if isinstance(value, montepy.input_parser.mcnp_input.Jump):
                jump_counter += 1
            else:
                flush_jumps()
                ret.append(value)
        flush_jumps()
        return ret

    @property
    def words(self):  # pragma: no cover
        """
        The words from the input file for this card.

        .. warning::
            .. deprecated:: 0.2.0
                This has been replaced by the syntax tree data structure.

        :raises DeprecationWarning: Access the syntax tree instead.
        """
        raise DeprecationWarning("This has been removed; instead use the syntax tree")

    @property
    def allowed_keywords(self):  # pragma: no cover
        """
        The allowed keywords for this class of MCNP_Card.

        The allowed keywords that would appear in the parameters block.
        For instance for cells the keywords ``IMP`` and ``VOL`` are allowed.
        The allowed keywords need to be in upper case.

        .. deprecated:: 0.2.0
            This is no longer needed. Instead this is specified in
            :func:`montepy.input_parser.tokens.MCNP_Lexer._KEYWORDS`.

        :returns: A set of the allowed keywords. If there are none this should return the empty set.
        :rtype: set
        """
        warnings.warn(
            "allowed_keywords are deprecated, and will be removed soon.",
            DeprecationWarning,
            stacklevel=2,
        )
        return set()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum


class Shortcuts(Enum):
    """
    Enumeration of the possible MCNP shortcuts.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.
    """

    REPEAT = "r"
    """
    A repeated entry shortcut.
    """
    JUMP = "j"
    """
    A jump entry, which counts as a default entry.
    """
    INTERPOLATE = "i"
    """
    A linear interpolation.
    """
    LOG_INTERPOLATE = "ilog"
    """
    A logarithmic interpolation.
    """
    MULTIPLY = "m"
    """
    a multiplication of the previous entry.
    """



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import montepy
from montepy.data_inputs.mode import Mode
from montepy.input_parser.mcnp_input import Input
from montepy.input_parser.block_type import BlockType
from montepy.particle import Particle


class TestMode(TestCase):
    def test_mode_init(self):
        in_str = "mode n p"
        mode = Mode(Input([in_str], BlockType.CELL))
        self.assertEqual(len(mode), 2)
        self.assertIn(Particle.NEUTRON, mode)
        self.assertIn(Particle.PHOTON, mode.particles)
        # test bad input
        in_str = "kcode"
        with self.assertRaises(montepy.errors.MalformedInputError):
            mode = Mode(Input([in_str], BlockType.CELL))
        in_str = "mode 1"
        with self.assertRaises(TypeError):
            mode = Mode(Input([in_str], BlockType.CELL))
        mode = Mode()
        self.assertEqual(len(mode), 1)
        self.assertIn(Particle.NEUTRON, mode)

    def test_mode_add(self):
        mode = Mode()
        with self.assertRaises(TypeError):
            mode.add(5)
        with self.assertRaises(ValueError):
            mode.add("5")
        mode.add("p")
        self.assertIn(Particle.PHOTON, mode)
        mode.add(Particle.NEGATIVE_MUON)
        self.assertIn(Particle.NEGATIVE_MUON, mode)
        self.assertEqual(len(mode), 3)

    def test_mode_remove(self):
        mode = Mode()
        with self.assertRaises(TypeError):
            mode.remove(5)
        with self.assertRaises(ValueError):
            mode.remove("5")
        mode.remove("n")
        self.assertEqual(len(mode), 0)
        self.assertTrue(Particle.NEUTRON not in mode)
        mode.add("p")
        # force update of syntax tree
        output = mode.format_for_mcnp_input((6, 2, 0))
        mode.remove(Particle.PHOTON)
        self.assertEqual(len(mode), 0)
        self.assertTrue(Particle.PHOTON not in mode)
        output = mode.format_for_mcnp_input((6, 2, 0))
        self.assertNotIn("p", output)
        self.assertNotIn("P", output)

    def test_mode_iter(self):
        mode = Mode()
        mode.add("p")
        parts = [Particle.NEUTRON, Particle.PHOTON]
        i = 0
        for particle in mode:
            self.assertIn(particle, parts)
            i += 1
        self.assertEqual(i, 2)

    def test_mode_format_input(self):
        mode = Mode()
        mode.add("e")
        mode.add("p")
        output = mode.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertEqual(len(output), 1)
        self.assertIn("MODE", output[0])
        self.assertIn("N", output[0])
        self.assertIn("P", output[0])
        self.assertIn("E", output[0])

    def test_set_mode(self):
        particles = {Particle.ELECTRON, Particle.PHOTON}
        mode = Mode()
        mode.set("e p")
        self.assertEqual(len(mode), 2)
        for part in particles:
            self.assertIn(part, mode)
        mode = Mode()
        mode.set(["e", "p"])
        self.assertEqual(len(mode), 2)
        for part in particles:
            self.assertIn(part, mode)
        mode = Mode()
        mode.set(particles)
        self.assertEqual(len(mode), 2)
        for part in particles:
            self.assertIn(part, mode)
        with self.assertRaises(TypeError):
            mode.set(5)
        with self.assertRaises(TypeError):
            mode.set([5])
        with self.assertRaises(ValueError):
            mode.set(["n", Particle.PHOTON])
        with self.assertRaises(ValueError):
            mode.set([Particle.PHOTON, "n"])



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import montepy

from montepy.errors import MalformedInputError
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input
from montepy.surfaces.axis_plane import AxisPlane
from montepy.surfaces.cylinder_on_axis import CylinderOnAxis
from montepy.surfaces.cylinder_par_axis import CylinderParAxis
from montepy.surfaces.general_plane import GeneralPlane
from montepy.surfaces.surface import Surface
from montepy.surfaces.surface_builder import surface_builder
from montepy.surfaces.surface_type import SurfaceType


class testSurfaces(TestCase):
    def test_surface_init(self):
        in_str = "1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        self.assertEqual(surf.number, 1)
        self.assertEqual(surf.old_number, 1)
        self.assertEqual(len(surf.surface_constants), 1)
        self.assertEqual(surf.surface_constants[0], 0.0)
        self.assertEqual(surf.surface_type, SurfaceType.PZ)
        self.assertFalse(surf.is_reflecting)
        self.assertFalse(surf.is_white_boundary)
        self.assertIsNone(surf.old_transform_number)
        self.assertIsNone(surf.transform)
        self.assertIsNone(surf.old_periodic_surface)
        self.assertIsNone(surf.periodic_surface)
        # test reflective
        in_str = "*1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        self.assertTrue(surf.is_reflecting)
        # test white boundary
        in_str = "+1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        self.assertTrue(surf.is_white_boundary)
        # test negative surface
        with self.assertRaises(MalformedInputError):
            in_str = "-1 PZ 0.0"
            card = Input([in_str], BlockType.SURFACE)
            Surface(card)
        # test bad surface number
        with self.assertRaises(MalformedInputError):
            in_str = "foo PZ 0.0"
            card = Input([in_str], BlockType.SURFACE)
            Surface(card)

        # test bad surface type
        with self.assertRaises(MalformedInputError):
            in_str = "1 INL 0.0"
            card = Input([in_str], BlockType.SURFACE)
            Surface(card)

        # test transform
        in_str = "1 5 PZ 0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        self.assertEqual(surf.old_transform_number, 5)

        # test periodic surface
        in_str = "1 -5 PZ 0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        self.assertEqual(surf.old_periodic_surface, 5)

        # test transform bad
        in_str = "1 5foo PZ 0"
        card = Input([in_str], BlockType.SURFACE)
        with self.assertRaises(MalformedInputError):
            Surface(card)
        in_str = "+1 PZ foo"
        card = Input([in_str], BlockType.SURFACE)
        with self.assertRaises(MalformedInputError):
            Surface(card)

    def test_validator(self):
        surf = Surface()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.format_for_mcnp_input((6, 2, 0))
        # cylinder on axis
        surf = CylinderOnAxis()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.format_for_mcnp_input((6, 2, 0))
        surf._surface_type = SurfaceType.CX
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        # cylinder par axis
        surf = CylinderParAxis()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.format_for_mcnp_input((6, 2, 0))
        surf._surface_type = SurfaceType.C_X
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        surf.radius = 5.0
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        # axis plane
        surf = AxisPlane()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.format_for_mcnp_input((6, 2, 0))
        surf._surface_type = SurfaceType.PX
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        surf.location = 0.0
        # ensure that this doesn't raise an error
        surf.validate()
        # general plane
        surf = GeneralPlane()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            surf.format_for_mcnp_input((6, 2, 0))
        surf._surface_type = SurfaceType.P
        with self.assertRaises(montepy.errors.IllegalState):
            surf.validate()

    def test_surface_is_reflecting_setter(self):
        in_str = "1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        surf.is_reflecting = True
        self.assertTrue(surf.is_reflecting)
        with self.assertRaises(TypeError):
            surf.is_reflecting = 1

    def test_surface_is_white_bound_setter(self):
        in_str = "1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        surf.is_white_boundary = True
        self.assertTrue(surf.is_white_boundary)
        with self.assertRaises(TypeError):
            surf.is_white_boundary = 1

    def test_surface_constants_setter(self):
        in_str = "1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        surf.surface_constants = [10.0]
        self.assertEqual(surf.surface_constants[0], 10.0)
        with self.assertRaises(TypeError):
            surf.surface_constants = "foo"
        with self.assertRaises(ValueError):
            surf.surface_constants = [1, "foo"]

    def test_surface_number_setter(self):
        in_str = "1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        surf.number = 20
        self.assertEqual(surf.number, 20)
        with self.assertRaises(TypeError):
            surf.number = "foo"
        with self.assertRaises(ValueError):
            surf.number = -5

    def test_surface_ordering(self):
        in_str = "1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf1 = Surface(card)
        in_str = "5 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf2 = Surface(card)
        sort_list = sorted([surf2, surf1])
        self.assertEqual(sort_list[0], surf1)
        self.assertEqual(sort_list[1], surf2)

    def test_surface_format_for_mcnp(self):
        in_str = "+1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        surf.number = 2
        answer = "+2 PZ 0.0"
        self.assertEqual(surf.format_for_mcnp_input((6, 2, 0))[0], answer)
        in_str = "*1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        surf.number = 2
        answer = "*2 PZ 0.0"
        self.assertEqual(surf.format_for_mcnp_input((6, 2, 0))[0], answer)
        # test input mimicry
        in_str = "1 PZ 0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        surf.number = 2
        answer = "2 PZ 0"
        self.assertEqual(surf.format_for_mcnp_input((6, 2, 0))[0], answer)

    def test_surface_str(self):
        in_str = "+1 PZ 0.0"
        card = Input([in_str], BlockType.SURFACE)
        surf = Surface(card)
        self.assertEqual(str(surf), "SURFACE: 1, PZ")
        self.assertEqual(
            repr(surf),
            "SURFACE: 1, PZ, periodic surface: None, transform: None, constants: [0.0]",
        )

    def test_surface_builder(self):
        testers = [
            ("1 PZ 0.0", AxisPlane),
            ("2 Cx 10.0", CylinderOnAxis),
            ("3 C/Z 4 3 5", CylinderParAxis),
            ("6 p 1 2 3 4", GeneralPlane),
            ("7 so 5", Surface),
            ("10 C/x 25 0 -5", CylinderParAxis),
            ("11 c/Y 25 0 -5", CylinderParAxis),
            ("12 CY 3", CylinderOnAxis),
            ("13 cz 0", CylinderOnAxis),
            ("14 px 1.e-3", AxisPlane),
            ("15 PY .1", AxisPlane),
        ]
        for in_str, surf_plane in testers:
            card = Input([in_str], BlockType.SURFACE)
            self.assertIsInstance(surface_builder(card), surf_plane)

    def test_axis_plane_init(self):
        bad_inputs = ["1 P 0.0", "1 PZ 0.0 10.0"]
        for bad_input in bad_inputs:
            with self.assertRaises(ValueError):
                surf = montepy.surfaces.axis_plane.AxisPlane(
                    Input([bad_input], BlockType.SURFACE)
                )

    def test_cylinder_on_axis_init(self):
        bad_inputs = ["1 P 0.0", "1 CZ 0.0 10.0"]
        for bad_input in bad_inputs:
            with self.assertRaises(ValueError):
                surf = montepy.surfaces.cylinder_on_axis.CylinderOnAxis(
                    Input([bad_input], BlockType.SURFACE)
                )

    def test_cylinder_par_axis_init(self):
        bad_inputs = ["1 P 0.0", "1 C/Z 0.0"]
        for bad_input in bad_inputs:
            with self.assertRaises(ValueError):
                surf = montepy.surfaces.cylinder_par_axis.CylinderParAxis(
                    Input([bad_input], BlockType.SURFACE)
                )

    def test_gen_plane_init(self):
        bad_inputs = ["1 PZ 0.0", "1 P 0.0"]
        for bad_input in bad_inputs:
            with self.assertRaises(ValueError):
                surf = montepy.surfaces.general_plane.GeneralPlane(
                    Input([bad_input], BlockType.SURFACE)
                )

    def test_axis_plane_location_setter(self):
        in_str = "1 PZ 0.0"
        surf = surface_builder(Input([in_str], BlockType.SURFACE))
        self.assertEqual(surf.location, 0.0)
        surf.location = 10.0
        self.assertEqual(surf.location, 10.0)
        with self.assertRaises(TypeError):
            surf.location = "hi"

    def test_cylinder_axis_radius_setter(self):
        in_str = "1 CZ 5.0"
        surf = surface_builder(Input([in_str], BlockType.SURFACE))
        self.assertEqual(surf.radius, 5.0)
        surf.radius = 3.0
        self.assertEqual(surf.radius, 3.0)
        with self.assertRaises(TypeError):
            surf.radius = "foo"
        with self.assertRaises(ValueError):
            surf.radius = -5.0

    def test_cylinder_radius_setter(self):
        in_str = "1 c/Z 3.0 4.0 5"
        surf = surface_builder(Input([in_str], BlockType.SURFACE))
        self.assertEqual(surf.radius, 5.0)
        surf.radius = 3.0
        self.assertEqual(surf.radius, 3.0)
        with self.assertRaises(TypeError):
            surf.radius = "foo"
        with self.assertRaises(ValueError):
            surf.radius = -5.0

    def test_cylinder_location_setter(self):
        in_str = "1 c/Z 3.0 4.0 5"
        surf = surface_builder(Input([in_str], BlockType.SURFACE))
        self.assertEqual(surf.coordinates, (3.0, 4.0))
        surf.coordinates = [1, 2]
        self.assertEqual(surf.coordinates, (1, 2))
        # test wrong type
        with self.assertRaises(TypeError):
            surf.coordinates = "fo"
        # test length issues
        with self.assertRaises(ValueError):
            surf.coordinates = [3, 4, 5]
